from opensnitch.plugins import PluginBase, PluginSignal


class Sample(PluginBase):
    # fields overriden from parent class
    name = "Sample"
    version = 0
    author = "opensnitch"
    created = ""
    modified = ""
    enabled = False
    description = "OpenSnitch sample plugin"

    # where this plugin be executed.
    TYPE = [PluginBase.TYPE_POPUPS]

    def __init__(self):
        self.signal_in.connect(self.cb_signal)

    def configure(self):
        pass

    def load_conf(self):
        pass

    def compile(self):
        """Transform a json object to python objects.
        """
        print("Sample.compile()")

    def run(self, args):
        """Run the action on the given arguments.
        """
        print("Sample.run() args:", args)

    def cb_signal(self, signal):
        print("Plugin.signal received:", self.name, signal)
        try:
            if signal['signal'] == PluginSignal.ENABLE:
                self.enabled = True
        except Exception as e:
            print("Plugin.Sample.cb_signal() exception:", e)



#!/usr/bin/env python3

#   Copyright (C) 2018      Simone Margaritelli
#                 2018      MiWCryptAnalytics
#                 2023      munix9
#                 2023      Wojtek Widomski
#                 2019-2025 Gustavo Iñiguez Goia
#
#   This file is part of OpenSnitch.
#
#   OpenSnitch is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   OpenSnitch is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with OpenSnitch.  If not, see <http://www.gnu.org/licenses/>.


import re

r = re.compile(r'(\d+)([smhdw]+)')

_second = 1
_minute = 60
_hour   = 60 * _minute
_day    = 60 * _hour
_week   = 60 * _day

_units = {
    's': _second,
    'm': _minute,
    'h': _hour,
    'd': _day,
    'w': _week
}

def to_seconds(dur_str):
    """converts a Golang duration string to seconds:
        "20s" -> 20 seconds
        "2m"  -> 120 seconds
        ...
    """
    secs = 0
    try:
        finds = r.findall(dur_str)
        for d in finds:
            try:
                unit = _units[d[1]]
                secs += (unit * int(d[0]))
            except:
                print("duration.to_seconds(): invalid unit:", d)

        return secs
    except:
        return secs



from PyQt5 import QtCore
from opensnitch.database import Database

class Config:
    __instance = None

    HELP_URL = "https://github.com/evilsocket/opensnitch/wiki/"
    HELP_RULES_URL = "https://github.com/evilsocket/opensnitch/wiki/Rules"
    HELP_SYS_RULES_URL = "https://github.com/evilsocket/opensnitch/wiki/System-rules#upgrading-from-previous-versions"
    HELP_SYSFW_URL = "https://github.com/evilsocket/opensnitch/wiki/System-rules"
    HELP_CONFIG_URL = "https://github.com/evilsocket/opensnitch/wiki/Configurations"
    HELP_SYSTRAY_WARN = "https://github.com/evilsocket/opensnitch/wiki/GUI-known-problems#gui-does-not-show-up"

    OPERAND_PROCESS_ID = "process.id"
    OPERAND_PROCESS_PATH = "process.path"
    OPERAND_PROCESS_COMMAND = "process.command"
    OPERAND_PROCESS_ENV = "process.env."
    OPERAND_PROCESS_HASH_MD5 = "process.hash.md5"
    OPERAND_PROCESS_HASH_SHA1 = "process.hash.sha1"
    OPERAND_USER_ID = "user.id"
    OPERAND_IFACE_OUT = "iface.out"
    OPERAND_IFACE_IN = "iface.in"
    OPERAND_SOURCE_IP = "source.ip"
    OPERAND_SOURCE_PORT = "source.port"
    OPERAND_DEST_IP = "dest.ip"
    OPERAND_DEST_HOST = "dest.host"
    OPERAND_DEST_PORT = "dest.port"
    OPERAND_DEST_NETWORK = "dest.network"
    OPERAND_SOURCE_NETWORK = "source.network"
    OPERAND_PROTOCOL = "protocol"
    OPERAND_LIST_DOMAINS = "lists.domains"
    OPERAND_LIST_DOMAINS_REGEXP = "lists.domains_regexp"
    OPERAND_LIST_IPS = "lists.ips"
    OPERAND_LIST_NETS = "lists.nets"

    RULE_TYPE_LIST = "list"
    RULE_TYPE_LISTS = "lists"
    RULE_TYPE_SIMPLE = "simple"
    RULE_TYPE_REGEXP = "regexp"
    RULE_TYPE_NETWORK = "network"
    RulesTypes = (RULE_TYPE_LIST, RULE_TYPE_LISTS, RULE_TYPE_SIMPLE, RULE_TYPE_REGEXP, RULE_TYPE_NETWORK)

    DEFAULT_TARGET_PROCESS = 0
    ACTION_DENY_IDX = 0
    ACTION_ALLOW_IDX = 1
    ACTION_REJECT_IDX = 2

    # don't translate
    ACTION_ALLOW = "allow"
    ACTION_DENY = "deny"
    ACTION_REJECT = "reject"
    ACTION_ACCEPT = "accept"
    ACTION_DROP = "drop"
    ACTION_JUMP = "jump"
    ACTION_REDIRECT = "redirect"
    ACTION_RETURN = "return"
    ACTION_TPROXY = "tproxy"
    ACTION_SNAT = "snat"
    ACTION_DNAT = "dnat"
    ACTION_MASQUERADE = "masquerade"
    ACTION_QUEUE = "queue"
    ACTION_LOG = "log"
    ACTION_STOP = "stop"

    DURATION_FIELD = "duration"
    DURATION_UNTIL_RESTART = "until restart"
    DURATION_ALWAYS = "always"
    DURATION_ONCE = "once"
    DURATION_12h = "12h"
    DURATION_1h = "1h"
    DURATION_30m = "30m"
    DURATION_15m = "15m"
    DURATION_5m = "5m"
    DURATION_30s = "30s"

    # Rules of this list are ignored/deleted
    RULES_DURATION_FILTER = ()
    # Rules of this list are active
    RULES_ACTIVE_TEMPORARY_RULES = ()
    RULES_TEMPORARY_LIST = [
        DURATION_ONCE, DURATION_30s, DURATION_5m,
        DURATION_15m, DURATION_30m, DURATION_1h,
        DURATION_12h,
        DURATION_UNTIL_RESTART]

    DEFAULT_DURATION_IDX = 6 # until restart

    POPUP_CENTER = 0
    POPUP_TOP_RIGHT = 1
    POPUP_BOTTOM_RIGHT = 2
    POPUP_TOP_LEFT = 3
    POPUP_BOTTOM_LEFT = 4

    DEFAULT_THEME = "global/theme"
    DEFAULT_THEME_DENSITY_SCALE = "global/theme_density_scale"
    DEFAULT_LANGUAGE = "global/language"
    DEFAULT_LANGNAME = "global/langname"
    DEFAULT_DISABLE_POPUPS = "global/disable_popups"
    DEFAULT_TIMEOUT_KEY  = "global/default_timeout"
    DEFAULT_ACTION_KEY   = "global/default_action"
    DEFAULT_DURATION_KEY = "global/default_duration"
    DEFAULT_TARGET_KEY   = "global/default_target"
    DEFAULT_IGNORE_RULES = "global/default_ignore_rules"
    DEFAULT_IGNORE_TEMPORARY_RULES = "global/default_ignore_temporary_rules"
    DEFAULT_POPUP_POSITION = "global/default_popup_position"
    DEFAULT_POPUP_ADVANCED = "global/default_popup_advanced"
    DEFAULT_POPUP_ADVANCED_DSTIP = "global/default_popup_advanced_dstip"
    DEFAULT_POPUP_ADVANCED_DSTPORT = "global/default_popup_advanced_dstport"
    DEFAULT_POPUP_ADVANCED_UID = "global/default_popup_advanced_uid"
    DEFAULT_POPUP_ADVANCED_CHECKSUM = "global/default_popup_advanced_checksum"
    DEFAULT_SERVER_ADDR  = "global/server_address"
    DEFAULT_SERVER_MAX_MESSAGE_LENGTH  = "global/server_max_message_length"
    DEFAULT_HIDE_SYSTRAY_WARN  = "global/hide_systray_warning"
    DEFAULT_DB_TYPE_KEY       = "database/type"
    DEFAULT_DB_FILE_KEY       = "database/file"
    DEFAULT_DB_PURGE_OLDEST   = "database/purge_oldest"
    DEFAULT_DB_MAX_DAYS       = "database/max_days"
    DEFAULT_DB_PURGE_INTERVAL = "database/purge_interval"
    DEFAULT_DB_JRNL_WAL       = "database/jrnl_wal"

    DEFAULT_TIMEOUT = 30

    NOTIFICATIONS_ENABLED = "notifications/enabled"
    NOTIFICATIONS_TYPE = "notifications/type"
    NOTIFICATION_TYPE_SYSTEM = 0
    NOTIFICATION_TYPE_QT = 1

    STATS_REFRESH_INTERVAL = "statsDialog/refresh_interval"
    STATS_GEOMETRY = "statsDialog/geometry"
    STATS_MAXIMIZED = "statsDialog/maximized"
    STATS_LAST_TAB = "statsDialog/last_tab"
    STATS_FILTER_TEXT = "statsDialog/general_filter_text"
    STATS_FILTER_ACTION = "statsDialog/general_filter_action"
    STATS_LIMIT_RESULTS = "statsDialog/general_limit_results"
    STATS_SHOW_COLUMNS = "statsDialog/show_columns"
    STATS_NODES_COL_STATE = "statsDialog/nodes_columns_state"
    STATS_GENERAL_COL_STATE = "statsDialog/general_columns_state"
    STATS_GENERAL_FILTER_TEXT = "statsDialog/"
    STATS_GENERAL_FILTER_ACTION = "statsDialog/"
    STATS_RULES_COL_STATE = "statsDialog/rules_columns_state"
    STATS_FW_COL_STATE = "statsDialog/firewall_columns_state"
    STATS_ALERTS_COL_STATE = "statsDialog/alerts_columns_state"
    STATS_NETSTAT_COL_STATE = "statsDialog/netstat_columns_state"
    STATS_RULES_TREE_EXPANDED_0 = "statsDialog/rules_tree_0_expanded"
    STATS_RULES_TREE_EXPANDED_1 = "statsDialog/rules_tree_1_expanded"
    STATS_RULES_SPLITTER_POS = "statsDialog/rules_splitter_pos"
    STATS_NODES_SPLITTER_POS = "statsDialog/nodes_splitter_pos"
    STATS_VIEW_COL_STATE =  "statsDialog/view_columns_state"
    STATS_VIEW_DETAILS_COL_STATE =  "statsDialog/view_details_columns_state"
    STATS_NETSTAT_FILTER_PROTO = "statsDialog/netstat_proto_filter"
    STATS_NETSTAT_FILTER_FAMILY = "statsDialog/netstat_family_filter"
    STATS_NETSTAT_FILTER_STATE = "statsDialog/netstat_state_filter"

    QT_PLATFORM_PLUGIN = "global/qt_platform_plugin"
    QT_AUTO_SCREEN_SCALE_FACTOR = "global/screen_scale_factor_auto"
    QT_SCREEN_SCALE_FACTOR = "global/screen_scale_factor"

    INFOWIN_GEOMETRY = "infoWindow/geometry"

    PLUGINS = "plugins/list"

    AUTH_TYPE = "auth/type"
    AUTH_CA_CERT = "auth/cacert"
    AUTH_CERT = "auth/cert"
    AUTH_CERTKEY = "auth/certkey"
    # don't translate

    @staticmethod
    def init():
        Config.__instance = Config()
        return Config.__instance

    @staticmethod
    def get():
        if Config.__instance == None:
            Config._instance = Config()
        return Config.__instance

    def __init__(self):
        self.settings = QtCore.QSettings("opensnitch", "settings")

        if self.settings.value(self.DEFAULT_TIMEOUT_KEY) == None:
            self.setSettings(self.DEFAULT_TIMEOUT_KEY, self.DEFAULT_TIMEOUT)
        if self.settings.value(self.DEFAULT_ACTION_KEY) == None:
            self.setSettings(self.DEFAULT_ACTION_KEY, self.ACTION_DENY_IDX)
        if self.settings.value(self.DEFAULT_DURATION_KEY) == None:
            self.setSettings(self.DEFAULT_DURATION_KEY, self.DEFAULT_DURATION_IDX)
        if self.settings.value(self.DEFAULT_TARGET_KEY) == None:
            self.setSettings(self.DEFAULT_TARGET_KEY, self.DEFAULT_TARGET_PROCESS)
        if self.settings.value(self.DEFAULT_DB_TYPE_KEY) == None:
            self.setSettings(self.DEFAULT_DB_TYPE_KEY, Database.DB_TYPE_MEMORY)
            self.setSettings(self.DEFAULT_DB_FILE_KEY, Database.DB_IN_MEMORY)
            self.setSettings(self.DEFAULT_DB_JRNL_WAL, Database.DB_JRNL_WAL)

        self.setRulesDurationFilter(
            self.getBool(self.DEFAULT_IGNORE_RULES),
            self.getInt(self.DEFAULT_IGNORE_TEMPORARY_RULES)
        )

    def reload(self):
        self.settings = QtCore.QSettings("opensnitch", "settings")

    def hasKey(self, key):
        return self.settings.contains(key)

    def setSettings(self, path, value):
        self.settings.setValue(path, value)
        self.settings.sync()

    def getSettings(self, path):
        return self.settings.value(path)

    def getBool(self, path, default_value=False):
        return self.settings.value(path, type=bool, defaultValue=default_value)

    def getInt(self, path, default_value=0):
        try:
            return self.settings.value(path, type=int, defaultValue=default_value)
        except Exception:
            return default_value

    def getDefaultAction(self):
        _default_action = self.getInt(self.DEFAULT_ACTION_KEY)
        if _default_action == self.ACTION_ALLOW_IDX:
            return self.ACTION_ALLOW
        else:
            return self.ACTION_DENY

    def setRulesDurationFilter(self, ignore_temporary_rules=False, temp_rules=1):
        try:
            if ignore_temporary_rules:
                Config.RULES_DURATION_FILTER = [
                    Config.DURATION_ONCE, Config.DURATION_30s, Config.DURATION_5m,
                    Config.DURATION_15m, Config.DURATION_30m, Config.DURATION_1h,
                    Config.DURATION_12h,
                    Config.DURATION_UNTIL_RESTART]

                Config.RULES_DURATION_FILTER = [
                    rule for rule in Config.RULES_TEMPORARY_LIST
                    if Config.RULES_TEMPORARY_LIST.index(rule) < temp_rules
                ]
                Config.RULES_ACTIVE_TEMPORARY_RULES = [
                    rule for rule in Config.RULES_TEMPORARY_LIST
                    if Config.RULES_TEMPORARY_LIST.index(rule) >= temp_rules
                ]
                #print("Temp rules preserved (RULES_DURATION_FILTER):", Config.RULES_DURATION_FILTER)
                #print("Temp rules to delete (ACTIVE_TEMPORARY_RULES):", Config.RULES_ACTIVE_TEMPORARY_RULES)

            else:
                Config.RULES_DURATION_FILTER = []
        except Exception as e:
            print("setRulesDurationFilter() exception:", e)

    def getMaxMsgLength(self):
        """return maximum configured length for the gRPC channel.
        Default size is 4MB, but in some scenarios it's not enough.
        """
        maxmsglen = 4194304
        maxmsglencfg = self.getSettings(Config.DEFAULT_SERVER_MAX_MESSAGE_LENGTH)
        if maxmsglencfg == '4MiB':
            maxmsglen = 4194304
        elif maxmsglencfg == '8MiB':
            maxmsglen = 8388608
        elif maxmsglencfg == '16MiB':
            maxmsglen = 16777216

        print("gRPC Max Message Length:", maxmsglencfg)
        print("                  Bytes:", maxmsglen)

        return maxmsglen



from .enums import *

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

class Chains():

    def __init__(self, nodes):
        self._nodes = nodes

    def get(self):
        chains = {}
        for node in self._nodes.get_nodes():
            chains[node] = self.get_node_chains(node)
        return chains

    def get_node_chains(self, addr):
        node = self._nodes.get_node(addr)
        if node == None:
            return rules
        if not 'firewall' in node:
            return rules

        chains = []
        for c in node['firewall'].SystemRules:
            # Chains node does not exist on <= v1.5.x
            try:
                chains.append(c.Chains)
            except Exception:
                pass
        return chains

    def get_node_chains(self, addr):
        node = self._nodes.get_node(addr)
        if node == None:
            return rules
        if not 'firewall' in node:
            return rules

        chains = []
        for c in node['firewall'].SystemRules:
            # Chains node does not exist on <= v1.5.x
            try:
                chains.append(c.Chains)
            except Exception:
                pass
        return chains



    def get_policy(self, node_addr=None, hook=Hooks.INPUT.value, _type=ChainType.FILTER.value, family=Family.INET.value):
        fwcfg = self._nodes.get_node(node_addr)['firewall']
        for sdx, n in enumerate(fwcfg.SystemRules):
            for cdx, c in enumerate(n.Chains):
                if c.Hook.lower() == hook and c.Type.lower() == _type and c.Family.lower() == family:
                    return c.Policy

        return None

    def set_policy(self, node_addr, hook=Hooks.INPUT.value, _type=ChainType.FILTER.value, family=Family.INET.value, policy=Policy.DROP):
        fwcfg = self._nodes.get_node(node_addr)['firewall']
        for sdx, n in enumerate(fwcfg.SystemRules):
            for cdx, c in enumerate(n.Chains):
                # XXX: support only "inet" family (ipv4/ipv6)? or allow to
                # specify ipv4 OR/AND ipv6? some systems have ipv6 disabled
                if c.Hook.lower() == hook and c.Type.lower() == _type and c.Family.lower() == family:
                    fwcfg.SystemRules[sdx].Chains[cdx].Policy = policy

                    if wantedHook == Fw.Hooks.INPUT.value and wantedPolicy == Fw.Policy.DROP.value:
                        fwcfg.SystemRules[sdx].Chains[cdx].Rules.extend([rule.Rules[0]])
                        self._nodes.add_fw_config(node_addr, fwcfg)
                    return True
        return False


    @staticmethod
    def new(
        name="",
        table=Table.FILTER.value,
        family=Family.INET.value,
        ctype="",
        hook=Hooks.INPUT.value
    ):
        chain = ui_pb2.FwChain()
        chain.Name = name
        chain.Table = table
        chain.Family = family
        chain.Type = ctype
        chain.Hook = hook

        return chain

# man nft
# Table 6. Standard priority names, family and hook compatibility matrix
# Name     │ Value │ Families                   │ Hooks
# raw      │ -300  │ ip, ip6, inet              │ all
# mangle   │ -150  │ ip, ip6, inet              │ all
# dstnat   │ -100  │ ip, ip6, inet              │ prerouting
# filter   │ 0     │ ip, ip6, inet, arp, netdev │ all
# security │ 50    │ ip, ip6, inet              │ all
# srcnat   │ 100   │ ip, ip6, inet              │ postrouting
#

class ChainFilter(Chains):
    """
    ChainFilter returns a new chain of type filter.

    The name of the chain is the one listed with: nft list table inet filter.
    It corresponds with the hook name, but can be a random name.
    """

    @staticmethod
    def input(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.INPUT.value
        chain.Table = Table.FILTER.value
        chain.Family = family
        chain.Type = ChainType.FILTER.value
        chain.Hook = Hooks.INPUT.value

        return chain

    @staticmethod
    def output(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.OUTPUT.value
        chain.Table = Table.FILTER.value
        chain.Family = family
        chain.Type = ChainType.FILTER.value
        chain.Hook = Hooks.OUTPUT.value

        return chain

    @staticmethod
    def forward(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.FORWARD.value
        chain.Table = Table.FILTER.value
        chain.Family = family
        chain.Type = ChainType.FILTER.value
        chain.Hook = Hooks.FORWARD.value

        return chain



class ChainMangle(Chains):
    """
    ChainMangle returns a new chain of type mangle.

    The name of the chain is the one listed with: nft list table inet mangle.
    It corresponds with the hook name, but can be a random name.
    """

    @staticmethod
    def output(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.OUTPUT.value
        chain.Table = Table.MANGLE.value

        chain.Family = family
        chain.Type = ChainType.MANGLE.value
        chain.Hook = Hooks.OUTPUT.value

        return chain

    @staticmethod
    def input(family=Family.INET.value):
        chain = ui_pb2.FwChain(family=Family.INET.value)
        chain.Name = Hooks.INPUT.value
        chain.Table = Table.MANGLE.value

        chain.Family = family
        chain.Type = ChainType.MANGLE.value
        chain.Hook = Hooks.INPUT.value

        return chain

    @staticmethod
    def forward(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.FORWARD.value
        chain.Table = Table.MANGLE.value

        chain.Family = family
        chain.Type = ChainType.MANGLE.value
        chain.Hook = Hooks.FORWARD.value

        return chain


    @staticmethod
    def prerouting(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.PREROUTING.value
        chain.Table = Table.MANGLE.value

        chain.Family = family
        chain.Type = ChainType.MANGLE.value
        chain.Hook = Hooks.PREROUTING.value

        return chain

    @staticmethod
    def postrouting(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.POSTROUTING.value
        chain.Table = Table.MANGLE.value

        chain.Family = family
        chain.Type = ChainType.MANGLE.value
        chain.Hook = Hooks.POSTROUTING.value

        return chain

class ChainDstNAT(Chains):
    """
    ChainDstNAT returns a new chain of type dstnat.

    The name of the chain is the one listed with: nft list table inet nat.
    It corresponds with the hook name, but can be a random name.
    """

    @staticmethod
    def prerouting(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.PREROUTING.value
        chain.Table = Table.NAT.value

        chain.Family = family
        chain.Type = ChainType.DNAT.value
        chain.Hook = Hooks.PREROUTING.value

        return chain

    @staticmethod
    def output(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.OUTPUT.value
        chain.Table = Table.NAT.value

        chain.Family = family
        chain.Type = ChainType.DNAT.value
        chain.Hook = Hooks.OUTPUT.value

        return chain

    @staticmethod
    def postrouting(family=Family.INET.value):
        chain = ui_pb2.FwChain()
        chain.Name = Hooks.POSTROUTING.value
        chain.Table = Table.NAT.value

        chain.Family = family
        chain.Type = ChainType.SNAT.value
        chain.Hook = Hooks.POSTROUTING.value

        return chain



from PyQt5 import QtCore
from abc import ABC, abstractmethod
from importlib import util
import os
import gc
import weakref

from opensnitch.config import Config

class PluginSignal(QtCore.QObject):
    signal = QtCore.pyqtSignal(dict)

    # actions to send to the plugins
    DISABLE = 0
    ENABLE = 1
    CONFIG_UPDATE = 2
    ERROR = 3
    STOP = 4

    def __init__(self):
        QtCore.QObject.__init__(self)

    def emit(self, args):
        self.signal.emit(args)

    def connect(self, callback):
        self.signal.connect(callback)

    def disconnect(self, callback):
        self.signal.disconnect(callback)

    #@QtCore.pyqtSlot(dict)
    def cb_signal(self, args):
        self.signal.disconnect(callback)

class PluginsList():
    """plugins store. Whenever a plugin is instantiated, it's added to the
    plugin list automatically
    """
    actions = []
    names = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.name != 'PluginBase' and cls.name not in PluginsList.names:
            cls.actions.append(cls)
            cls.names[cls.name] = cls

class PluginBase(PluginsList, ABC):
    """Base class for every plugin.
    A plugin may be applied on different areas of the GUI:
        - globally
          * Background tasks.
          * periodic downloaders.
          * DB.
        - popups
        - views (generic connections lists)
        - view-details
        - procs-list (processes listed on the Proc tab)
        - domains-list (domains listed on the Host tab)
        - proc-dialog

    It may have support for one or multiple types.

    Every plugin can connect to the signal 'updated', to receive events, like:
        - enable/disable
        - configuration update/reload.

    As of v1.7.0 plugins are not called directly. They're defined in "actions",
    compiled when loading, and configured when called from every view, dialog,
    etc.

    When calling compile(), every plugin must create the python objects needed,
    so they can be reused later when run() is called.

    When calling configure(), every plugin is responsable to modify the GUI
    as needed, adding new buttons, modifying existing widgets, behaviour, ...
    """
    name = "PluginBase"
    version = 0
    author = "opensnitch"
    created = ""
    modified = ""
    description = "<empty description>"
    #enabled = False

    TYPE = []
    # allowed types that every plugin must declare
    TYPE_GLOBAL = "global"
    TYPE_POPUPS = "popups"
    TYPE_MAIN_DIALOG = "main-dialog"
    TYPE_PROC_DIALOG = "proc-dialog"
    TYPE_PROCS_LIST = "procs-list"
    TYPE_DOMAINS_LIST = "domains-list"
    TYPE_IPS_LIST = "ips-list"
    TYPE_VIEWS = "views"
    TYPE_VIEW_DETAILS = "view-details"

    # Generic signal to send data to all plugins' instances.
    # A plugin may have been instanstiated on many configurations,
    # this way we can enable/update/etc all the instances of a particular
    # plugin.

    # signals sent to the plugin
    signal_in = PluginSignal()

    # signals emitted by the plugin
    signal_out = PluginSignal()

    def __init__(self):
        pass

    def get_name(self):
        return self.name

    def is_enabled(self):
        return self.enabled

    def set_enabled(self, enable):
        self.enabled = enable

    def get_description(self):
        return self.description

    @abstractmethod
    def configure(self):
        raise NotImplementedError("Needs to be implemented")

    @abstractmethod
    def compile(self):
        raise NotImplementedError("Needs to be implemented")

    @abstractmethod
    def run(self, args):
        raise NotImplementedError("Needs to be implemented")

class PluginsManager():
    __instance = None

    @staticmethod
    def instance():
        if PluginsManager.__instance == None:
            PluginsManager.__instance = PluginsManager()
        return PluginsManager.__instance

    def __init__(self):
        path = os.path.abspath(__file__)
        # TODO: allow to load plugins from a different location
        self._plugins_loaded = weakref.WeakValueDictionary()
        self._plugins_path = os.path.dirname(path)

        self._cfg = Config.get()
        self.enabled_plugins = self._cfg.getSettings(Config.PLUGINS)
        # FIXME: don't hardcode this plugin here.
        # For now, load Highlight plugin by default.
        if self.enabled_plugins == None:
            self.enabled_plugins = ['highlight']
        # if there's only 1 plugin enabled, the type will be str instead of
        # list
        if type(self.enabled_plugins) == str:
            self.enabled_plugins = [self.enabled_plugins]
        if 'highlight' not in self.enabled_plugins:
            self.enabled_plugins.append('highlight')
        #print("enabled plugins >>", self.enabled_plugins)

    def load_plugins(self):
        #print("PluginsManager.load_plugins()", self._plugins_path)
        for dname in os.listdir(self._plugins_path):
            self.load_plugin_byname(dname)

        #print("PluginsManager:", PluginsList.actions)
        for plug in PluginsList.actions:
            p = plug()
            p.signal_in.emit({"plugin": p.get_name(), "signal": PluginSignal.ENABLE})
            #print("plugin ->", p.get_name())

    def load_plugin_byname(self, name, force=False):
        path = os.path.join(self._plugins_path, name)
        if os.path.isdir(path) == False:
            return False
        # the loading of the plugin is based on the file name
        pname = os.path.join(path, name + ".py")
        #print("load_plugin_byname.plugin path:", pname)
        if os.path.exists(pname) == False:
            return False

        return self.load_plugin(pname, force)

    def load_plugin(self, path, force=False):
        """loads the .py file, the name of the file is used to load the plugin
        """
        #print("loading plugin:", path)
        name = os.path.split(path)[-1]
        if self.enabled_plugins == None and not force:
            #print("skipping not enabled plugin:", name)
            return False
        if name[:-3] not in self.enabled_plugins and force == False:
            #print("PluginsManager: plugin disabled:", name[:-3])
            return False

        # Whenever a plugin is loaded, it's automatically added to the
        # PluginsList class.
        spec = util.spec_from_file_location(name, path)
        module = util.module_from_spec(spec)
        spec.loader.exec_module(module)

        self._plugins_loaded[module] = spec

        return True

    def unload_all(self):
        for mod in self._plugins_loaded:
            #print("PluginsManager.unload_all()", mod)
            del mod
        gc.collect()



from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtCore import QCoreApplication as QC
from google.protobuf.json_format import MessageToJson
import uuid

from .enums import Operator
from .exprs import ExprLog

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

class Rules(QObject):
    rulesUpdated = pyqtSignal()

    # Fields defined in the protobuf, to be used as constants on other parts.
    FIELD_UUID = "UUID"
    FIELD_ENABLED = "Enabled"
    FIELD_TARGET = "Target"

    def __init__(self, nodes):
        QObject.__init__(self)
        self._nodes = nodes
        self.rulesUpdated.connect(self._cb_rules_updated)

    def _cb_rules_updated(self):
        pass

    def add(self, addr, rule):
        """Add a new rule to the corresponding table on the given node
        """
        node = self._nodes.get_node(addr)
        if node == None or not 'firewall' in node:
            return False, QC.translate("firewall", "rule not found by its ID.")
        if self.is_duplicated(addr, rule):
            return False, QC.translate("firewall", "duplicated.")

        for sdx, n in enumerate(node['firewall'].SystemRules):
            for cdx, c in enumerate(n.Chains):
                if c.Name == rule.Name and \
                        c.Hook == rule.Hook and \
                        c.Table == rule.Table and \
                        c.Family == rule.Family and \
                        c.Type == rule.Type:
                    node['firewall'].SystemRules[sdx].Chains[cdx].Rules.extend([rule.Rules[0]])
                    node['fwrules'][rule.Rules[0].UUID] = rule
                    self._nodes.add_fw_config(addr, node['firewall'])
                    self._nodes.add_fw_rules(addr, node['fwrules'])

                    self.rulesUpdated.emit()
                    return True

        return False, QC.translate("firewall", "firewall table/chain not properly configured.")

    def insert(self, addr, rule, position=0):
        """Insert a new rule to the corresponding table on the given node
        """
        node = self._nodes.get_node(addr)
        if node == None or not 'firewall' in node:
            return False, QC.translate("firewall", "this node doesn't have a firewall configuration, review it.")
        if self.is_duplicated(addr, rule):
            return False, QC.translate("firewall", "duplicated")

        for sdx, n in enumerate(node['firewall'].SystemRules):
            for cdx, c in enumerate(n.Chains):
                if c.Name == rule.Name and \
                        c.Hook == rule.Hook and \
                        c.Table == rule.Table and \
                        c.Family == rule.Family and \
                        c.Type == rule.Type:
                    if hasattr(node['firewall'].SystemRules[sdx].Chains[cdx].Rules, "insert"):
                        node['firewall'].SystemRules[sdx].Chains[cdx].Rules.insert(int(position), rule.Rules[0])
                    else:
                        node['firewall'].SystemRules[sdx].Chains[cdx].Rules.extend([rule.Rules[0]])
                    node['fwrules'][rule.Rules[0].UUID] = rule
                    self._nodes.add_fw_config(addr, node['firewall'])
                    self._nodes.add_fw_rules(addr, node['fwrules'])

                    self.rulesUpdated.emit()
                    return True, ""

        return False, QC.translate("firewall", "firewall table/chain not properly configured.")


    def update(self, addr, uuid, rule):
        node = self._nodes.get_node(addr)
        if node == None or not 'firewall' in node:
            return False, QC.translate("firewall", "this node doesn't have a firewall configuration, review it.")
        for sdx, n in enumerate(node['firewall'].SystemRules):
            for cdx, c in enumerate(n.Chains):
                for rdx, r in enumerate(c.Rules):
                    if r.UUID == uuid:
                        c.Rules[rdx].CopyFrom(rule.Rules[0])
                        node['firewall'].SystemRules[sdx].Chains[cdx].Rules[rdx].CopyFrom(rule.Rules[0])
                        self._nodes.add_fw_config(addr, node['firewall'])
                        node['fwrules'][uuid] = rule
                        self._nodes.add_fw_rules(addr, node['fwrules'])

                        self.rulesUpdated.emit()
                        return True, ""

        return False, QC.translate("firewall", "rule not found by its ID.")

    def get(self):
        rules = []
        for node in self._nodes.get_nodes():
            node_rules = self.get_by_node(node)
            rules += node_rules

        return rules

    def delete(self, addr, uuid):
        node = self._nodes.get_node(addr)
        if node == None or not 'firewall' in node:
            return False, None
        for sdx, n in enumerate(node['firewall'].SystemRules):
            for cdx, c in enumerate(n.Chains):
                for idx, r in enumerate(c.Rules):
                    if r.UUID == uuid:
                        del node['firewall'].SystemRules[sdx].Chains[cdx].Rules[idx]
                        self._nodes.add_fw_config(addr, node['firewall'])
                        if uuid in node['fwrules']:
                            del node['fwrules'][uuid]
                            self._nodes.add_fw_rules(addr, node['fwrules'])
                        else:
                            # raise Error("rules doesn't have UUID field")
                            return False, None

                        self.rulesUpdated.emit()
                        return True, node['firewall']

        return False, None

    def get_by_node(self, addr):
        rules = []
        node = self._nodes.get_node(addr)
        if node == None:
            return rules
        if not 'firewall' in node:
            return rules
        for u in node['firewall'].SystemRules:
            for c in u.Chains:
                for r in c.Rules:
                    rules.append(Rules.to_array(addr, c, r))
        return rules

    def get_by_uuid(self, addr, uuid):
        rules = []
        node = self._nodes.get_node(addr)
        if node == None:
            return rules
        if not 'firewall' in node:
            return rules
        for u in node['firewall'].SystemRules:
            for c in u.Chains:
                for r in c.Rules:
                    if r.UUID == uuid:
                        return r
        return None

    def swap(self, view, addr, uuid, old_pos, new_pos):
        """
        swap changes the order of 2 rows.

        The list of rules is ordered from top to bottom: 0,1,2,3...
        so a click on the down button sums +1, a click on the up button rest -1
        """
        node = self._nodes.get_node(addr)
        if node == None:
            return
        if not 'firewall' in node:
            return
        for sdx, c in enumerate(node['firewall'].SystemRules):
            for cdx, u in enumerate(c.Chains):
                nrules = len(u.Rules)
                for rdx, r in enumerate(u.Rules):
                    # is the last rule
                    if new_pos > nrules and new_pos < nrules:
                        break
                    if u.Rules[rdx].UUID == uuid:
                        old_rule = u.Rules[old_pos]
                        new_rule = ui_pb2.FwRule()
                        new_rule.CopyFrom(u.Rules[new_pos])

                        node['firewall'].SystemRules[sdx].Chains[cdx].Rules[new_pos].CopyFrom(old_rule)
                        node['firewall'].SystemRules[sdx].Chains[cdx].Rules[old_pos].CopyFrom(new_rule)

                        self._nodes.add_fw_config(addr, node['firewall'])
                        #self._nodes.add_fw_rules(addr, node['fwrules'])

                        self.rulesUpdated.emit()
                        return True
        return False

    def is_duplicated(self, addr, orig_rule):
        # we need to duplicate the rule, otherwise we'd modify the UUID of the
        # orig rule.
        temp_c = ui_pb2.FwChain()
        temp_c.CopyFrom(orig_rule)
        # the UUID will be different, so zero it out.
        # but keep a copy of the original one.
        orig_uuid = temp_c.Rules[0].UUID
        temp_c.Rules[0].UUID = ""
        node = self._nodes.get_node(addr)
        if node == None:
            return False
        if not 'firewall' in node:
            return False
        for n in node['firewall'].SystemRules:
            for c in n.Chains:
                if c.Name == temp_c.Name and \
                        c.Hook == temp_c.Hook and \
                        c.Table == temp_c.Table and \
                        c.Family == temp_c.Family and \
                        c.Type == temp_c.Type:
                    for rdx, r in enumerate(c.Rules):
                        uuid = c.Rules[rdx].UUID
                        c.Rules[rdx].UUID = ""
                        is_equal = (c.Rules[rdx].SerializeToString() == temp_c.Rules[0].SerializeToString() or orig_uuid == uuid)
                        c.Rules[rdx].UUID = uuid

                        if is_equal:
                            return True

        return False

    @staticmethod
    def new(
            enabled=True,
            _uuid="",
            description="",
            expressions=None,
            target="",
            target_parms=""
            ):
        rule = ui_pb2.FwRule()
        if _uuid == "":
            rule.UUID = str(uuid.uuid1())
        else:
            rule.UUID = _uuid
        rule.Enabled = enabled
        rule.Description = description
        if expressions != None:
            rule.Expressions.extend([expressions])
        rule.Target = target
        rule.TargetParameters = target_parms

        return rule

    @staticmethod
    def new_flat(c, r):
        """Create a new "flat" rule from a hierarchical one.
        Transform from:
            {
             xx:
                 {
                   yy: {
        to:
            {xx:, yy}
        """

        chain = ui_pb2.FwChain()
        chain.CopyFrom(c)
        del chain.Rules[:]
        chain.Rules.extend([r])

        return chain

    @staticmethod
    def to_dict(sysRules):
        """Transform json/protobuf struct to flat structure.
        This is the default format used to find rules in the table view.
        """
        rules={}
        for s in sysRules:
            for c in s.Chains:
                if len(c.Rules) == 0:
                    continue
                for r in c.Rules:
                    rules[r.UUID] = Rules.new_flat(c, r)

        return rules

    @staticmethod
    def to_json(rule):
        try:
            return MessageToJson(rule)
        except:
            return None

    @staticmethod
    def to_array(addr, chain, rule):
        cols = []
        cols.append(rule.UUID)
        cols.append(addr)
        cols.append(chain.Name)
        cols.append(chain.Table)
        cols.append(chain.Family)
        cols.append(chain.Hook)
        cols.append(str(rule.Enabled))
        cols.append(rule.Description)
        exprs = ""
        for e in rule.Expressions:
            exprs += "{0} {1}".format(
                e.Statement.Name,
                "".join(
                    [
                        "{0} {1}{2} ".format(
                            h.Key,
                            e.Statement.Op + " " if e.Statement.Op != Operator.EQUAL.value else "",
                            "\"{0}\"".format(h.Value) if h.Key == ExprLog.PREFIX.value else h.Value
                        ) for h in e.Statement.Values
                    ]
                )
            )
        cols.append(exprs)
        cols.append(rule.Target)
        cols.append(rule.TargetParameters)

        return cols



import os
import sys
import json
import re

from PyQt5 import QtCore, QtGui, uic, QtWidgets

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

from opensnitch.nodes import Nodes
from opensnitch.desktop_parser import LinuxDesktopParser
from opensnitch.utils import Message, Icons
from opensnitch.actions import Actions
from opensnitch.plugins import PluginBase
from opensnitch.config import Config

DIALOG_UI_PATH = "%s/../res/process_details.ui" % os.path.dirname(sys.modules[__name__].__file__)
class ProcessDetailsDialog(QtWidgets.QDialog, uic.loadUiType(DIALOG_UI_PATH)[0]):

    LOG_TAG = "[ProcessDetails]: "

    _notification_callback = QtCore.pyqtSignal(str, ui_pb2.NotificationReply)

    TAB_STATUS          = 0
    TAB_DESCRIPTORS     = 1
    TAB_IOSTATS         = 2
    TAB_MAPS            = 3
    TAB_STACK           = 4
    TAB_ENVS            = 5

    TABS = {
            TAB_STATUS: {
                "text": None,
                "scrollPos": 0
                },
            TAB_DESCRIPTORS: {
                "text": None,
                "scrollPos": 0
                },
            TAB_IOSTATS: {
                "text": None,
                "scrollPos": 0
                },
            TAB_MAPS: {
                "text": None,
                "scrollPos": 0
                },
            TAB_STACK: {
                "text": None,
                "scrollPos": 0
                },
            TAB_ENVS: {
                "text": None,
                "scrollPos": 0
                }
            }

    SOCKET_REGEX = "(socket.*:.*state.*)"

    def __init__(self, parent=None, appicon=None):
        super(ProcessDetailsDialog, self).__init__(parent)
        QtWidgets.QDialog.__init__(self, parent, QtCore.Qt.WindowStaysOnTopHint)
        self.setWindowFlags(QtCore.Qt.Window)
        self.setupUi(self)
        self.setWindowIcon(appicon)

        self._app_name = None
        self._app_icon = None
        self._apps_parser = LinuxDesktopParser()
        self._nodes = Nodes.instance()
        self._notification_callback.connect(self._cb_notification_callback)
        self._actions = Actions.instance()
        self._action_list = self._actions.getByType(PluginBase.TYPE_PROC_DIALOG)
        self._configure_plugins()

        self._nid = None
        self._pid = ""
        self._notifications_sent = {}

        self._socketsRE = re.compile(self.SOCKET_REGEX)

        self.cmdClose.clicked.connect(self._cb_close_clicked)
        self.cmdAction.clicked.connect(self._cb_action_clicked)
        self.comboPids.currentIndexChanged.connect(self._cb_combo_pids_changed)

        self.TABS[self.TAB_STATUS]['text'] = self.textStatus
        self.TABS[self.TAB_DESCRIPTORS]['text'] = self.textOpenedFiles
        self.TABS[self.TAB_IOSTATS]['text'] = self.textIOStats
        self.TABS[self.TAB_MAPS]['text'] = self.textMappedFiles
        self.TABS[self.TAB_STACK]['text'] = self.textStack
        self.TABS[self.TAB_ENVS]['text'] = self.textEnv

        self.TABS[self.TAB_DESCRIPTORS]['text'].setFont(QtGui.QFont("monospace"))

        self.iconStart = QtGui.QIcon.fromTheme("media-playback-start")
        self.iconPause = QtGui.QIcon.fromTheme("media-playback-pause")

        closeIcon = Icons.new(self, "window-close")
        self.cmdClose.setIcon(closeIcon)
        self.iconStart = Icons.new(self, "media-playback-start")
        self.iconPause = Icons.new(self, "media-playback-pause")

    def _configure_plugins(self):
        for conf in self._action_list:
            action = self._action_list[conf]
            for name in action['actions']:
                try:
                    action['actions'][name].configure(self)
                except Exception as e:
                    print("procdialog._configure_plugins() exception:", name, " you may want to enable this plugin -", e)

    @QtCore.pyqtSlot(str, ui_pb2.NotificationReply)
    def _cb_notification_callback(self, addr, reply):
        if reply.id not in self._notifications_sent:
            print("[stats] unknown notification received: ", reply.id)
        else:
            noti = self._notifications_sent[reply.id]

            if reply.code == ui_pb2.ERROR:
                self._show_message(
                    QtCore.QCoreApplication.translate(
                        "proc_details",
                        "<b>Error loading process information:</b> <br><br>\n\n") + reply.data
                )
                self._pid = ""
                self._set_button_running(False)

                # if we haven't loaded any data yet, just close the window
                if self._data_loaded == False:
                    # but if there're more than 1 pid keep the window open.
                    # we may have one pid already closed and one alive.
                    if self.comboPids.count() <= 1:
                        self._close()

                self._delete_notification(reply.id)
                return

            if noti.type != ui_pb2.TASK_START:
                print("proc-details, invalid notification type?", noti)
                return

            if noti.type == ui_pb2.TASK_START and reply.data != "":
                self._load_data(reply.data)

            elif noti.type == ui_pb2.TASK_STOP:
                if reply.data != "":
                    self._show_message(
                        QtCore.QCoreApplication.translate(
                            "proc_details",
                            "<b>Error stopping monitoring process:</b><br><br>") + reply.data
                    )
                    self._set_button_running(False)

                self._delete_notification(reply.id)

    def closeEvent(self, e):
        self._close()

    def _cb_close_clicked(self):
        self._close()

    def _cb_combo_pids_changed(self, idx):
        if idx == -1:
            return
        # TODO: this event causes to send to 2 Start notifications
        #if self._pid != "" and self._pid != self.comboPids.currentText():
        #    self._stop_monitoring()
        #    self._pid = self.comboPids.currentText()
        #    self._start_monitoring()

    def _cb_action_clicked(self):
        if not self.cmdAction.isChecked():
            self._stop_monitoring()
        else:
            self._start_monitoring()

    def _show_message(self, text):
        Message.ok(text, "", QtWidgets.QMessageBox.Warning)

    def _delete_notification(self, nid):
        if nid in self._notifications_sent:
            del self._notifications_sent[nid]

    def _reset(self):
        self._app_name = None
        self._app_icon = None
        self.comboPids.clear()
        self.labelProcName.setText(QtCore.QCoreApplication.translate("proc_details", "loading..."))
        self.labelProcArgs.setText(QtCore.QCoreApplication.translate("proc_details", "loading..."))
        self.labelProcPath.setText(QtCore.QCoreApplication.translate("proc_details", "loading..."))
        self.labelProcIcon.clear()
        self.labelStatm.setText("")
        self.labelCwd.setText("")
        self.labelChecksums.setText("")
        self.labelParent.setText("")
        for tidx in range(0, len(self.TABS)):
            self.TABS[tidx]['text'].setPlainText("")

    def _set_button_running(self, yes):
        if yes:
            self.cmdAction.setChecked(True)
            self.cmdAction.setIcon(self.iconPause)
        else:
            self.cmdAction.setChecked(False)
            self.cmdAction.setIcon(self.iconStart)

    def _close(self):
        self._stop_monitoring()
        self.comboPids.clear()
        self._pid = ""
        self.hide()

    def monitor(self, pids):
        if self._pid != "":
            self._stop_monitoring()

        self._data_loaded = False
        self._pids = pids
        self._reset()
        for pid in pids:
            if pid != None:
                self.comboPids.addItem(pid)

        self.show()
        self._start_monitoring()

    def _set_tab_text(self, tab_idx, text):
        self.TABS[tab_idx]['scrollPos'] = self.TABS[tab_idx]['text'].verticalScrollBar().value()
        self.TABS[tab_idx]['text'].setPlainText(text)
        self.TABS[tab_idx]['text'].verticalScrollBar().setValue(self.TABS[tab_idx]['scrollPos'])

    def _start_monitoring(self):
        try:
            # avoid to send notifications without a pid
            if self._pid != "":
                return

            self._pid = self.comboPids.currentText()
            if self._pid == "":
                return

            self._set_button_running(True)
            noti = ui_pb2.Notification(
                clientName="",
                serverName="",
                type=ui_pb2.TASK_START,
                data=self._build_notification_message(self._pid),
                rules=[])
            self._nid = self._nodes.send_notification(self._pids[self._pid], noti, self._notification_callback)
            self._notifications_sent[self._nid] = noti
        except Exception as e:
            print(self.LOG_TAG + "exception starting monitoring: ", e)

    def _stop_monitoring(self):
        if self._pid == "":
            return

        self._set_button_running(False)
        noti = ui_pb2.Notification(
            clientName="",
            serverName="",
            type=ui_pb2.TASK_STOP,
            data=self._build_notification_message(self._pid),
            rules=[])
        self._nid = self._nodes.send_notification(self._pids[self._pid], noti, self._notification_callback)
        self._notifications_sent[self._nid] = noti
        self._pid = ""
        self._app_icon = None

    def _load_data(self, data):
        """Load the process information received via notifications.
        """
        tab_idx = self.tabWidget.currentIndex()

        try:
            proc = json.loads(data)
            self._load_app_icon(proc['Path'])
            if self._app_name != None:
                self.labelProcName.setText("<b>" + self._app_name + "</b>")
                self.labelProcName.setToolTip("<b>" + self._app_name + "</b>")

            if 'Tree' in proc:
                proc['Tree'].reverse()
                self.labelParent.setText(
                    "<b>Parent(s): </b>" + " \U000027a1 ".join(
                        path['key'] for path in proc['Tree']
                    )
                )
            else:
                self.labelParent.setText("<could not obtain hash>")

            if proc['Path'] not in proc['Args']:
                self.labelProcPath.setVisible(True)
                self.labelProcPath.setText("({0})".format(proc['Path']))
            else:
                self.labelProcPath.setVisible(False)

            if 'Checksums' in proc:
                checksums = proc['Checksums']
                hashes = ""
                if Config.OPERAND_PROCESS_HASH_MD5 in checksums:
                    hashes = "<b>md5:</b> {0}".format(checksums[Config.OPERAND_PROCESS_HASH_MD5])
                if Config.OPERAND_PROCESS_HASH_SHA1 in checksums:
                    hashes = "<b>sha1:</b> {0}".format(checksums[Config.OPERAND_PROCESS_HASH_SHA1])

                self.labelChecksums.setText(hashes)


            self.labelProcArgs.setFixedHeight(30)
            self.labelProcArgs.setText(" ".join(proc['Args']))
            self.labelProcArgs.setToolTip(" ".join(proc['Args']))
            self.labelCwd.setText("<b>CWD: </b>" + proc['CWD'])
            self.labelCwd.setToolTip("<b>CWD: </b>" + proc['CWD'])
            self._load_mem_data(proc['Statm'])

            if tab_idx == self.TAB_STATUS:
                self._set_tab_text(tab_idx, proc['Status'])

            elif tab_idx == self.TAB_DESCRIPTORS:
                self._load_descriptors(proc['Descriptors'])

            elif tab_idx == self.TAB_IOSTATS:
                self._load_iostats(proc['IOStats'])

            elif tab_idx == self.TAB_MAPS:
                self._set_tab_text(tab_idx, proc['Maps'])

            elif tab_idx == self.TAB_STACK:
                self._set_tab_text(tab_idx, proc['Stack'])

            elif tab_idx == self.TAB_ENVS:
                self._load_env_vars(proc['Env'])

            self._data_loaded = True

        except Exception as e:
            print(self.LOG_TAG + "exception loading data: ", e)

    def _load_app_icon(self, proc_path):
        if self._app_icon != None:
            return

        self._app_name, self._app_icon, _, _ = self._apps_parser.get_info_by_path(proc_path, "terminal")
        pixmap = Icons.get_by_appname(self._app_icon)
        self.labelProcIcon.setPixmap(pixmap)

        if self._app_name == None:
            self._app_name = proc_path

    def _load_iostats(self, iostats):
        ioText = "%-16s %dMB<br>%-16s %dMB<br>%-16s %d<br>%-16s %d<br>%-16s %dMB<br>%-16s %dMB<br>" % (
                "<b>Chars read:</b>",
                ((iostats['RChar'] / 1024) / 1024),
                "<b>Chars written:</b>",
                ((iostats['WChar'] / 1024) / 1024),
                "<b>Syscalls read:</b>",
                (iostats['SyscallRead']),
                "<b>Syscalls write:</b>",
                (iostats['SyscallWrite']),
                "<b>KB read:</b>",
                ((iostats['ReadBytes'] / 1024) / 1024),
                "<b>KB written: </b>",
                ((iostats['WriteBytes'] / 1024) / 1024)
                )

        self.textIOStats.setPlainText("")
        self.textIOStats.appendHtml(ioText)

    def _load_mem_data(self, mem):
        memText = "<b>VIRT:</b> %dMB, <b>RSS:</b> %dMB, <b>Libs:</b> %dMB, <b>Data:</b> %dMB, <b>Text:</b> %dMB" % (
                (mem['Size'] / 1024) / 1024,
                (mem['Resident'] / 1024) / 1024,
                (mem['Lib'] / 1024) / 1024,
                (mem['Data'] / 1024) / 1024,
                (mem['Text'] / 1024) / 1024
                )
        self.labelStatm.setText(memText)

    def _load_descriptors(self, descriptors):
        text = "%-12s%-40s%-8s -> %s\n\n" % ("Size", "Time", "Name", "Symlink")
        for d in descriptors:
            if self.checkFilterFiles.isChecked() and d['Size'] == 0:
                continue
            text += "{:<12}{:<40}{:<8} -> {}\n".format(str(d['Size']), d['ModTime'], d['Name'], d['SymLink'])

        if self.checkFilterSockets.isChecked():
            matches = self._socketsRE.findall(text)
            if len(matches) > 0:
                sockets = ""
                for m in matches:
                    sockets = "{0}\n{1}".format(sockets, m)

                if self.checkFilterFiles.isChecked():
                    text += sockets
                else:
                    text = sockets

        self._set_tab_text(self.TAB_DESCRIPTORS, text)

    def _load_env_vars(self, envs):
        if envs == {}:
            self._set_tab_text(self.TAB_ENVS, "<no environment variables>")
            return

        text = "%-15s\t%s\n\n" % ("Name", "Value")
        for env_name in envs:
            text += "%-15s:\t%s\n" % (env_name, envs[env_name])

        self._set_tab_text(self.TAB_ENVS, text)

    def _build_notification_message(self, pid):
        # TODO: make interval configurable
        return '{"name": "pid-monitor", "data": { "interval": "5s", "pid": "%s" }}' % pid



import sys
import time
import os
import os.path
import json

from PyQt5 import QtCore, QtGui, uic, QtWidgets
from PyQt5.QtCore import QCoreApplication as QC

from opensnitch.utils import Icons, Message
from opensnitch.config import Config
from opensnitch.nodes import Nodes
from opensnitch.dialogs.firewall_rule import FwRuleDialog
import opensnitch.firewall as Fw
import opensnitch.firewall.profiles as FwProfiles

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

DIALOG_UI_PATH = "%s/../res/firewall.ui" % os.path.dirname(sys.modules[__name__].__file__)
class FirewallDialog(QtWidgets.QDialog, uic.loadUiType(DIALOG_UI_PATH)[0]):
    LOG_TAG = "[fw dialog]"

    COMBO_IN = 0
    COMBO_OUT = 1

    POLICY_ACCEPT = 0
    POLICY_DROP = 1

    _notification_callback = QtCore.pyqtSignal(str, ui_pb2.NotificationReply)

    def __init__(self, parent=None, appicon=None, node=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.setupUi(self)
        self.setWindowIcon(appicon)
        self.appicon = appicon

        # TODO: profiles are ready to be used. They need to be tested, and
        # create some default profiles (home, office, public, ...)
        self.comboProfile.setVisible(False)
        self.lblProfile.setVisible(False)

        self.secHighIcon = Icons.new(self, "security-high")
        self.secMediumIcon = Icons.new(self, "security-medium")
        self.secLowIcon = Icons.new(self, "security-low")
        self.lblStatusIcon.setPixmap(self.secHighIcon.pixmap(96, 96))

        self._fwrule_dialog = FwRuleDialog(appicon=self.appicon)
        self._cfg = Config.get()
        self._fw = Fw.Firewall.instance()
        self._nodes = Nodes.instance()
        self._fw_profiles = {}
        self._last_profile = {
            self.COMBO_IN: FwProfiles.ProfileAcceptInput.value,
            self.COMBO_OUT: FwProfiles.ProfileAcceptOutput.value
        }

        self._notification_callback.connect(self._cb_notification_callback)
        self._notifications_sent = {}

        self._nodes.nodesUpdated.connect(self._cb_nodes_updated)
        self.cmdNewRule.clicked.connect(self._cb_new_rule_clicked)
        self.cmdAllowOUTService.clicked.connect(self._cb_allow_out_service_clicked)
        self.cmdAllowINService.clicked.connect(self._cb_allow_in_service_clicked)
        self.comboInput.currentIndexChanged.connect(lambda: self._cb_combo_policy_changed(self.COMBO_IN))
        self.comboProfile.currentIndexChanged.connect(self._cb_combo_profile_changed)
        self.sliderFwEnable.valueChanged.connect(self._cb_enable_fw_changed)
        self.cmdClose.clicked.connect(self._cb_close_clicked)
        self.cmdHelp.clicked.connect(
            lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl(Config.HELP_SYSFW_URL))
        )

        # TODO: when output policy is set to Drop, all outbound traffic is
        # blocked.
        self.comboOutput.currentIndexChanged.connect(lambda: self._cb_combo_policy_changed(self.COMBO_OUT))

        closeIcon = Icons.new(self, "window-close")
        excludeIcon = Icons.new(self, "go-up")
        allowInIcon = Icons.new(self, "go-down")
        newIcon = Icons.new(self, "document-new")
        helpIcon = Icons.new(self, "help-browser")
        self.cmdClose.setIcon(closeIcon)
        self.cmdAllowOUTService.setIcon(excludeIcon)
        self.cmdAllowINService.setIcon(allowInIcon)
        self.cmdNewRule.setIcon(newIcon)
        self.cmdHelp.setIcon(helpIcon)

    @QtCore.pyqtSlot(str, ui_pb2.NotificationReply)
    def _cb_notification_callback(self, addr, reply):
        self.comboInput.setEnabled(True)
        self.comboOutput.setEnabled(True)
        if reply.id in self._notifications_sent:
            if reply.code == ui_pb2.OK:
                self._set_status_successful(QC.translate("firewall", "Configuration applied."))

            else:
                self._set_status_error(QC.translate("firewall", "There was an error: {0}").format(reply.data))

            del self._notifications_sent[reply.id]
        else:
            print(self.LOG_TAG, "unknown notification:", reply)


    @QtCore.pyqtSlot(int)
    def _cb_nodes_updated(self, total):
        self._check_fw_status()

    def _cb_combo_profile_changed(self, idx):
        combo_profile = self._fw_profiles[idx]
        json_profile = json.dumps(list(combo_profile.values())[0]['Profile'])

        for addr in self._nodes.get():
            node = self._nodes.get_node(addr)
            if 'firewall' not in node:
                print("ERROR: node {0} has no 'firewall'".format(node))
                continue
            fwcfg = node['firewall']
            ok, err = self._fw.apply_profile(addr, json_profile)
            if ok:
                self.send_notification(addr, fwcfg)
            else:
                self._set_status_error(QC.translate("firewall", "error adding profile extra rules:", err))

    def _cb_combo_policy_changed(self, combo):
        self._reset_status_message()
        self.comboInput.setEnabled(False)
        self.comboOutput.setEnabled(False)

        wantedProfile = FwProfiles.ProfileAcceptInput.value
        if combo == self.COMBO_OUT:
            wantedProfile = FwProfiles.ProfileAcceptOutput.value
            if self.comboOutput.currentIndex() == self.POLICY_DROP:
                wantedProfile = FwProfiles.ProfileDropOutput.value
        else:
            if self.comboInput.currentIndex() == self.POLICY_DROP:
                wantedProfile = FwProfiles.ProfileDropInput.value


        if combo == self.COMBO_IN and \
                self.comboInput.currentIndex() == self.POLICY_ACCEPT:
            json_profile = json.dumps(FwProfiles.ProfileDropInput.value)
            for addr in self._nodes.get():
                fwcfg = self._nodes.get_node(addr)['firewall']
                ok, err = self._fw.delete_profile(addr, json_profile)
                if not ok:
                    print(err)
        elif combo == self.COMBO_OUT and \
                self.comboOutput.currentIndex() == self.POLICY_ACCEPT:
                #self._last_profile[self.COMBO_IN] == FwProfiles.ProfileDropInput.value and \
            self._set_status_message(
                QC.translate("firewall", "Warning: Output policy configured to drop. If OpenSnitch dies, outbound network traffic will be blocked.")
            )
            json_profile = json.dumps(FwProfiles.ProfileDropOutput.value)
            for addr in self._nodes.get():
                fwcfg = self._nodes.get_node(addr)['firewall']
                ok, err = self._fw.delete_profile(addr, json_profile)
                #if ok:
                #    self.send_notification(addr, fwcfg)


        json_profile = json.dumps(wantedProfile)
        for addr in self._nodes.get():
            fwcfg = self._nodes.get_node(addr)['firewall']
            ok, err = self._fw.apply_profile(addr, json_profile)
            if ok:
                self.send_notification(addr, fwcfg)
            else:
                self._set_status_error(QC.translate("firewall", "Policy not applied: {0}".format(err)))

        self._last_profile[combo] = wantedProfile

    def _cb_new_rule_clicked(self):
        self.new_rule()

    def _cb_allow_out_service_clicked(self):
        self.allow_out_service()

    def _cb_allow_in_service_clicked(self):
        self.allow_in_service()

    def _cb_enable_fw_changed(self, enable):
        if self._nodes.count() == 0:
            self.sliderFwEnable.blockSignals(True)
            self.sliderFwEnable.setValue(False)
            self.sliderFwEnable.blockSignals(False)
            return
        self.enable_fw(enable)

    def _cb_close_clicked(self):
        self._close()

    def _load_nodes(self):
        self._nodes = self._nodes.get()

    def _close(self):
        self.hide()

    def _change_fw_backend(self, addr, node_cfg):
        nid, notif = self._nodes.change_node_config(addr, node_cfg, self._notification_callback)
        self._notifications_sent[nid] = notif

    def showEvent(self, event):
        super(FirewallDialog, self).showEvent(event)
        self._reset_fields()
        self._check_fw_status()
        self._fw_profiles = FwProfiles.Profiles.load_predefined_profiles()
        self.comboProfile.blockSignals(True)
        for pr in self._fw_profiles:
            self.comboProfile.addItem([pr[k] for k in pr][0]['Name'])
        self.comboProfile.blockSignals(False)

    def send_notification(self, node_addr, fw_config):
        self._set_status_message(QC.translate("firewall", "Applying changes..."))
        nid, notif = self._nodes.reload_fw(node_addr, fw_config, self._notification_callback)
        self._notifications_sent[nid] = {'addr': node_addr, 'notif': notif}

    def _check_fw_status(self):
        self.lblFwStatus.setText("")
        self.sliderFwEnable.blockSignals(True)
        self.comboInput.blockSignals(True)
        self.comboOutput.blockSignals(True)
        self.comboProfile.blockSignals(True)

        self._disable_widgets()

        try:
            enableFw = False
            enableFwBtn = (self._nodes.count() > 0)
            self.sliderFwEnable.setEnabled(enableFwBtn)
            if not enableFwBtn:
                return

            # TODO: handle nodes' firewall properly
            for addr in self._nodes.get():
                node = self._nodes.get_node(addr)
                self._fwConfig = node['firewall']
                enableFw |= self._fwConfig.Enabled

                if self.fw_is_incompatible(addr, node):
                    enableFw = False
                    return

                # XXX: Here we loop twice over the chains. We could have 1 loop.
                pol_in = self._fw.chains.get_policy(addr, Fw.Hooks.INPUT.value)
                pol_out = self._fw.chains.get_policy(addr, Fw.Hooks.OUTPUT.value, Fw.ChainType.MANGLE.value)

                if pol_in != None:
                    self.comboInput.setCurrentIndex(
                        Fw.Policy.values().index(pol_in)
                    )
                else:
                    self._set_status_error(QC.translate("firewall", "Error getting INPUT chain policy"))
                    self._disable_widgets()
                if pol_out != None:
                    self.comboOutput.setCurrentIndex(
                        Fw.Policy.values().index(pol_out)
                    )
                else:
                    self._set_status_error(QC.translate("firewall", "Error getting OUTPUT chain policy"))
                    self._disable_widgets()

        except Exception as e:
            self._set_status_error("Firewall status error (report on github please): {0}".format(e))

        finally:
            # some nodes may have the firewall disabled whilst other enabled
            #if not enableFw:
            #    self.lblFwStatus(QC.translate("firewall", "Some nodes have the firewall disabled"))

            self._disable_widgets(not enableFw)
            self.lblStatusIcon.setEnabled(enableFw)
            self.sliderFwEnable.setValue(enableFw)

            self.sliderFwEnable.blockSignals(False)
            self.comboInput.blockSignals(False)
            self.comboOutput.blockSignals(False)
            self.comboProfile.blockSignals(False)

    def fw_is_incompatible(self, addr, node):
        """Check if the fw is compatible with this GUI.
        If it's incompatible, disable the option to enable it.
        """
        incompatible = False
        # firewall iptables is not supported from the GUI.
        # display a warning
        node_cfg = json.loads(node['data'].config)
        if node_cfg['Firewall'] == "iptables":
            self._disable_widgets()
            self.sliderFwEnable.setEnabled(False)
            if self.isHidden() == False and self.change_fw(addr, node_cfg):
                    node_cfg['Firewall'] = "nftables"
                    self.sliderFwEnable.setEnabled(True)
                    self.enable_fw(True)
                    self._change_fw_backend(addr, node_cfg)
                    return False
            incompatible = True

        if node['data'].systemFirewall.Version == 0:
            self._disable_widgets()
            self.sliderFwEnable.setEnabled(False)
            self.lblFwStatus.setText(
                QC.translate("firewall", "<html>The firewall configuration is outdated,\n"
                            "you need to update it to the new format: <a href=\"{0}\">learn more</a>"
                            "</html>".format(Config.HELP_SYS_RULES_URL)
            ))
            incompatible = True

        return incompatible

    def change_fw(self, addr, node_cfg):
        """Ask the user to change fw iptables to nftables
        """
        ret = Message.yes_no(
            QC.translate("firewall",
                        "In order to configure firewall rules from the GUI, we need to use 'nftables' instead of 'iptables'"
                        ),
            QC.translate("firewall", "Change default firewall to 'nftables' on node {0}?".format(addr)),
            QtWidgets.QMessageBox.Warning)
        if ret != QtWidgets.QMessageBox.Cancel:
            return True

        return False

    def enable_fw(self, enable):
        try:
            self._disable_widgets(not enable)
            if enable:
                self._set_status_message(QC.translate("firewall", "Enabling firewall..."))
            else:
                self._set_status_message(QC.translate("firewall", "Disabling firewall..."))

            # if previous input policy was DROP, when disabling the firewall it
            # must be ACCEPT to allow output traffic.
            if not enable and self.comboInput.currentIndex() == self.POLICY_DROP:
                self.comboInput.blockSignals(True)
                self.comboInput.setCurrentIndex(self.POLICY_ACCEPT)
                self.comboInput.blockSignals(False)
                for addr in self._nodes.get():
                    json_profile = json.dumps(FwProfiles.ProfileAcceptInput.value)
                    ok, err = self._fw.apply_profile(addr, json_profile)
                    if not ok:
                        self._set_status_error(
                            QC.translate("firewall", "Error applying INPUT ACCEPT profile: {0}".format(err))
                        )
                        return

            if not enable and self.comboOutput.currentIndex() == self.POLICY_DROP:
                self.comboOutput.blockSignals(True)
                self.comboOutput.setCurrentIndex(self.POLICY_ACCEPT)
                self.comboOutput.blockSignals(False)
                for addr in self._nodes.get():
                    json_profile = json.dumps(FwProfiles.ProfileAcceptOutput.value)
                    ok, err = self._fw.apply_profile(addr, json_profile)
                    if not ok:
                        self._set_status_error(
                            QC.translate("firewall", "Error applying OUTPUT ACCEPT profile: {0}".format(err))
                        )
                        return

            for addr in self._nodes.get():
                # FIXME:
                # Due to how the daemon reacts to events when the fw configuration
                # is modified, changing the policy + disabling the fw doesn't work
                # as expected.
                # The daemon detects that the fw is disabled, and it never changes
                # the policy.
                # As a workaround to this problem, we send 2 fw changes:
                # - one for changing the policy
                # - another one for disabling the fw

                fwcfg = self._nodes.get_node(addr)['firewall']
                self.send_notification(addr, fwcfg)
                time.sleep(0.5)
                fwcfg.Enabled = True if enable else False
                self.send_notification(addr, fwcfg)

            self.lblStatusIcon.setEnabled(enable)
            self.policiesBox.setEnabled(enable)

            time.sleep(0.5)

        except Exception as e:
            QC.translate("firewall", "Error: {0}".format(e))

    def load_rule(self, addr, uuid):
        self._fwrule_dialog.load(addr, uuid)

    def new_rule(self):
        self._fwrule_dialog.new()

    def allow_out_service(self):
        self._fwrule_dialog.exclude_service(self.COMBO_OUT)

    def allow_in_service(self):
        self._fwrule_dialog.exclude_service(self.COMBO_IN)

    def _set_status_error(self, msg):
        self.statusLabel.show()
        self.statusLabel.setStyleSheet('color: red')
        self.statusLabel.setText(msg)

    def _set_status_successful(self, msg):
        self.statusLabel.show()
        self.statusLabel.setStyleSheet('color: green')
        self.statusLabel.setText(msg)

    def _set_status_message(self, msg):
        self.statusLabel.show()
        self.statusLabel.setStyleSheet('color: darkorange')
        self.statusLabel.setText(msg)

    def _reset_status_message(self):
        self.statusLabel.setText("")
        self.statusLabel.hide()

    def _reset_fields(self):
        self._reset_status_message()

    def _disable_widgets(self, disable=True):
        self.comboInput.setEnabled(not disable)
        self.comboOutput.setEnabled(not disable)
        self.cmdNewRule.setEnabled(not disable)
        self.cmdAllowOUTService.setEnabled(not disable)
        self.cmdAllowINService.setEnabled(not disable)




from opensnitch.nodes import Nodes
from opensnitch.database import Database
from opensnitch.database.enums import ConnFields
from opensnitch.utils import Utils
from opensnitch.utils.infowindow import InfoWindow
from PyQt5.QtCore import QCoreApplication as QC

class ConnDetails(InfoWindow):
    """Display a small dialog with the details of a connection
    """

    def __init__(self, parent):
        super().__init__(parent)

        self._db = Database.instance()
        self._nodes = Nodes.instance()

    def showByField(self, field, value):
        records = self._db.get_connection_by_field(field, value)
        if not records.next():
            return

        node = records.value(ConnFields.Node)
        uid = records.value(ConnFields.UID)
        if self._nodes.is_local(node):
            uid = Utils.get_user_id(uid)

        conn_text = QC.translate("stats", """
<b>{0}</b><br><br>
<b>Time:</b> {1}<br><br>
<b>Process:</b><br>{2}<br>
<b>Cmdline:</b><br>{3}<br>
<b>CWD:</b><br>{4}<br><br>
<b>UID:</b> {5} <b>PID:</b> {6}<br>
<br>
<b>Node:</b> {7}<br><br>
<b>{8}</b> {9}:{10} -> {11} ({12}):{13}
<br><br>
<b>Rule:</b><br>
{14}
""".format(
                    records.value(ConnFields.Action).upper(),
                    records.value(ConnFields.Time),
                    records.value(ConnFields.Process),
                    records.value(ConnFields.Cmdline),
                    records.value(ConnFields.CWD),
                    uid,
                    records.value(ConnFields.PID),
                    node,
                    records.value(ConnFields.Protocol).upper(),
                    records.value(ConnFields.SrcPort),
                    records.value(ConnFields.SrcIP),
                    records.value(ConnFields.DstIP),
                    records.value(ConnFields.DstHost),
                    records.value(ConnFields.DstPort),
                    records.value(ConnFields.Rule)
                ))

        self.showText(conn_text)



import json
import os
import glob
import sys

from PyQt5.QtCore import QObject

from opensnitch.utils.xdg import xdg_config_home
from opensnitch.actions.default_configs import (
    commonDelegateConfig,
    rulesDelegateConfig,
    fwDelegateConfig,
    netstatDelegateConfig
)

from opensnitch.plugins import PluginsList
from opensnitch.plugins import PluginsManager

class Actions(QObject):
    """List of actions to perform on the data that is displayed on the GUI,
    defined in JSON files.

    Whenever an item (popup, cell, etc) matches a condition, an action
    (config/plugin) is applied. For example:
        - if the text of a cell matches a condition for the given columns,
        then the properties of the cell/row and the text are colorized.
        - if the result of an analysis of a domain is malicious, colorize
        popups' text labels in red + add a tab with the result of the analysis.

    The actions are defined in JSON format:
        {
        "created": "....",
        "name": "...",
        "type": ["views"],
        "actions": {
            "highlight": {
                "cells": [
                    {
                        "text": ["allow", "True", "online"],
                        "cols": [3,5,6],
                        "color": "green",
                    },
                    {
                        "text": ["deny", "False", "offline"],
                        "cols": [3,5,6],
                        "color": "red",
                    }
                ],
                "rows": []
            }
        }

        "type" field is the area of the GUI where the actions will be applied:

             "global" -> global actions, like background tasks.
             "views"  -> applies to Views (lists of items), QItemDelegate
             "popups" -> applies to popups.
             "main-dialog" -> applies to the main window.
             "proc-dialog" -> applies to the Process dialog.
             "procs-list" -> applies to the Procs view. (TODO)
             "domains-list" -> applies to the Domains view. (TODO)
             "ips-list" -> applies to the IPs view. (TODO)
             "db" -> applies to the DB. Modify items before inserting, react to
              data being added, etc. (TODO)

        "actions" is the list of actions to execute:

            - the name of the action defines the python plugin to load:
                "highligh" -> plugins/highligh/highlight.py
                "downloader" -> plugins/downloader/downloader.py, etc.
            - every action has its own plugin (*.py file) which is in charge
            of parse and compile to configuration if needed.
              For example for "highlight" action, "color": "red" is compiled
              to QtColor("red")

    There're 3 hardcoded actions by default of type Highlight:
        - rules: applied to the rules to colorize the columns Enabled and
        Action
        - firewall: applied to the fw rules to colorize the columns Action and
        Target.
        - common: applied to the rest of the views to colorize the column
        Action.

    Users can modify the default actions, by adding more patterns to colorize,
    and saving them to $XDG_CONFIG_HOME/opensnitch/actions/myaction.json

    At the same time they can also create new actions in that directory
    to be applied on certain views.

    """
    __instance = None

    # list of loaded actions
    _actions_list = {}
    _plugins = []


    KEY_ACTIONS = "actions"
    KEY_NAME = "name"
    KEY_TYPE = "type"

    # TODO: emit a signal when the actions are (re)loaded
    # reloaded_signal = pyQtSignal()

    # default paths to look for actions
    _paths = [
        #os.path.dirname(sys.modules[__name__].__file__) + "/data/",
        "{0}/{1}".format(xdg_config_home, "/opensnitch/actions/")
    ]

    @staticmethod
    def instance():
        if Actions.__instance == None:
            Actions.__instance = Actions()
        return Actions.__instance

    def __init__(self, parent=None):
        QObject.__init__(self)
        self._actions_list = {}
        self._plugin_mgr = PluginsManager.instance()
        try:
            base_dir = "{0}/{1}".format(xdg_config_home, "/opensnitch/actions/")
            os.makedirs(base_dir, 0o700)
        except Exception as e:
            print("actions.__init__ exception:", e)
        #print("ActionsLists:", PluginsList.actions)

    def _load_default_configs(self):
        # may be overwritten by user choice
        self._actions_list[commonDelegateConfig[Actions.KEY_NAME]] = self.compile(commonDelegateConfig)
        self._actions_list[rulesDelegateConfig[Actions.KEY_NAME]] = self.compile(rulesDelegateConfig)
        self._actions_list[fwDelegateConfig[Actions.KEY_NAME]] = self.compile(fwDelegateConfig)
        self._actions_list[netstatDelegateConfig[Actions.KEY_NAME]] = self.compile(netstatDelegateConfig)

    def load(self, action_file):
        """read a json file from disk and create the action."""
        try:
            with open(action_file, 'r') as fd:
                data=fd.read()
                obj = json.loads(data)
                action = self.compile(obj)
                return obj, action
        except:
            pass

        return None, None

    def loadAll(self):
        """look for actions firstly on default system path, secondly on
        user's home.
        If a user customizes existing configurations, they'll be saved under
        the user's home directory.

        Action files are .json files.
        """
        self._load_default_configs()

        for path in self._paths:
            for jfile in glob.glob(os.path.join(path, '*.json')):
                #print("Actions.loadconf()", jfile)
                obj, action = self.load(jfile)
                if obj and action:
                    self._actions_list[obj[Actions.KEY_NAME]] = action

    def compile(self, json_obj):
        """translates json definitions to python objects"""
        try:
            if Actions.KEY_NAME not in json_obj or json_obj[Actions.KEY_NAME] == "":
                return None
            if json_obj.get(Actions.KEY_ACTIONS) == None:
                return None

            # "actions": { "highlight": ..., "virustotal": ..., }
            #print("plugins >>", PluginsList.names)
            for action_name in json_obj[Actions.KEY_ACTIONS]:
                action_obj = json_obj[Actions.KEY_ACTIONS][action_name]
                if action_obj == None or action_obj.get('enabled') == None or action_obj.get('enabled') == False:
                    print("actions.compile() skipping disabled action '{0}'".format(action_name))
                    # FIXME: if one of the action is not enabled, we're
                    # invalidating all the configured actions.
                    return None

                # see if the plugin is loaded, if it's not, try to load it.
                if PluginsList.names.get(action_name.capitalize()) == None:
                    if self._plugin_mgr.load_plugin_byname(action_name, force=True) == False:
                        print("actions.compile() unable to load plugin name '{0}'".format(action_name))
                        return None

                # allow to use "Plugin" or "plugin" to name actions in json
                # files.
                # python class will be always capitalized.
                _p = PluginsList.names.get(action_name.capitalize())
                plug = _p(action_obj)

                # compile the plugin, preparing/configuring the plugin
                # before it's used.
                plug.compile()

                # save the "compiled" action to the action list
                json_obj[Actions.KEY_ACTIONS][action_name]= plug

            return json_obj
        except Exception as e:
            print("Actions.compile() exception:", e)
            return None

    def getAll(self):
        return self._actions_list

    def deleteAll(self):
        self._actions_list = {}

    def get(self, name):
        try:
            return self._actions_list[name]
        except Exception as e:
            print("actions.get() exception:", e, "name:", name)
            return None

    def getByType(self, acttype):
        try:
            actlist = {}
            for name in self._actions_list:
                act = self._actions_list[name]
                if act == None:
                    print("actions.getByType() none:", name)
                    continue
                types = act.get('type')
                if types == None:
                    continue
                if acttype in types:
                    actlist[name] = self._actions_list[name]
            return actlist
        except Exception as e:
            print("actions.getByType() ({0}) exception: {1}".format(acttype, e))
            return None

    def delete(self, name):
        try:
            del self._actions_list[name]
            # TODO:
            # self.reloaded_signal.emit()
        except:
            pass

    def isValid(self):
        pass



#   Copyright (C) 2024      Nolan Carouge
#
#   This file is part of OpenSnitch.
#
#   OpenSnitch is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   OpenSnitch is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with OpenSnitch.  If not, see <http://www.gnu.org/licenses/>.

import json
import ipaddress
import os

class NetworkAliases:
    ALIASES = {}

    @staticmethod
    def load_aliases():
        # Define the path to the network_aliases.json file
        script_dir = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(script_dir, 'network_aliases.json')

        # Check if the file exists before attempting to load it
        if not os.path.exists(filename):
            raise FileNotFoundError(f"The file '{filename}' does not exist.")

        # Load the JSON file
        with open(filename, 'r') as f:
            NetworkAliases.ALIASES = json.load(f)
        print(f"Loaded network aliases from {filename}")  # Confirmation message

    @staticmethod
    def get_alias(ip):
        try:
            ip_obj = ipaddress.ip_address(ip)
            for alias, networks in NetworkAliases.ALIASES.items():
                for network in networks:
                    net_obj = ipaddress.ip_network(network)
                    if ip_obj in net_obj:
                        return alias
        except ValueError:
            pass
        return None

    @staticmethod
    def get_networks_for_alias(alias):
        return NetworkAliases.ALIASES.get(alias, [])

    @staticmethod
    def get_alias_all():
        # Return a list of all alias names
        return list(NetworkAliases.ALIASES.keys())

# Load aliases at startup
try:
    NetworkAliases.load_aliases()
except FileNotFoundError as e:
    print(e)




from opensnitch.config import Config
from PyQt5 import QtCore, QtWidgets, QtGui

class InfoWindow(QtWidgets.QDialog):
    """Display a text on a small dialog.
    """
    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent, QtCore.Qt.Tool)
        self.setContentsMargins(0, 0, 0, 0)

        self._cfg = Config.get()

        self.layout = QtWidgets.QVBoxLayout(self)
        self._textedit = QtWidgets.QTextEdit()
        # hide cursor
        self._textedit.setCursorWidth(0)
        self._textedit.setViewportMargins(QtCore.QMargins(0,0,0,0))
        self._textedit.setMinimumSize(300, 325)
        self._textedit.setReadOnly(True)
        self._textedit.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse | QtCore.Qt.TextSelectableByKeyboard)
        self._textedit.setAutoFillBackground(True)
        self._textedit.setStyleSheet("QLabel { background: yellow }")

        self.layout.addWidget(self._textedit)

        self._load_settings()

    def closeEvent(self, ev):
        self._save_settings()
        ev.accept()
        self.hide()

    def _load_settings(self):
        saved_geometry = self._cfg.getSettings(Config.INFOWIN_GEOMETRY)
        if saved_geometry is not None:
            self.restoreGeometry(saved_geometry)

    def _save_settings(self):
        self._cfg.setSettings(Config.INFOWIN_GEOMETRY, self.saveGeometry())

    def showText(self, text):
        self._load_settings()

        self._textedit.setText(text)
        #self.resize(self.tooltip_textedit.sizeHint())

        pos = QtGui.QCursor.pos()
        win_size = self.size()
        # center dialog on cursor, relative to the parent widget.
        x_off = (int(win_size.width()/2))
        y_off = (int(win_size.height()/2))
        point = QtCore.QPoint(
            pos.x()-x_off, pos.y()-y_off
        )
        self.move(point.x(), point.y())

        self.show()

    def showHtml(self, text):
        self._load_settings()

        self._textedit.setHtml(text)

        pos = QtGui.QCursor.pos()
        win_size = self.size()
        # center dialog on cursor, relative to the parent widget.
        x_off = (int(win_size.width()/2))
        y_off = (int(win_size.height()/2))
        point = QtCore.QPoint(
            pos.x()-x_off, pos.y()-y_off
        )
        self.move(point.x(), point.y())

        self.show()



import json
from PyQt5 import QtWidgets, QtGui, QtCore
from opensnitch.utils import Icons
from opensnitch.plugins.virustotal import _utils

def build_vt_tab(plugin, parent):
    """add a new tab with a text field that will contain the result of the query in JSON format.
    """
    backIcon = Icons.new(plugin, "go-previous")

    # FIXME: find the widget with the name 'vt_tab', there could be more
    # plugins that are tabs.
    prev_wdg = parent.stackedWidget.widget(3)
    if prev_wdg != None and prev_wdg.objectName() == "vt_tab":
        return prev_wdg

    wdg = QtWidgets.QWidget()
    gridLayout =  QtWidgets.QGridLayout()
    hor_wdg = QtWidgets.QHBoxLayout()
    cmdBack = QtWidgets.QPushButton("", objectName="plugin_virustotal")
    cmdBack.setFlat(True)
    cmdBack.setIcon(backIcon)

    # 0 details, 1 checksums, 2 main
    cmdBack.clicked.connect(lambda: parent.stackedWidget.setCurrentIndex(2))
    cmdBack.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
    textWdg = QtWidgets.QTextEdit()
    textWdg.setTextInteractionFlags(
        QtCore.Qt.LinksAccessibleByMouse | QtCore.Qt.TextSelectableByKeyboard | QtCore.Qt.TextSelectableByMouse
    )
    textWdg.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
    wdg.setObjectName("vt_tab")
    gridLayout.setContentsMargins(5, 3, 5, 5)
    gridLayout.setVerticalSpacing(3)

    hor_wdg.addWidget(cmdBack)
    hor_wdg.addStretch(1)
    #hor_wdg.addWidget(spacer)
    gridLayout.addLayout(hor_wdg, 0, 0)
    gridLayout.addWidget(textWdg, 1, 0)
    wdg.setLayout(gridLayout)

    return wdg

def add_vt_tab(parent, tab):
    parent.stackedWidget.addWidget(tab)

def add_vt_response(parent, response, error=None):
    tab = parent.stackedWidget.widget(3).layout()
    textWdg = tab.itemAtPosition(1, 0).widget()
    textWdg.clear()
    #textWdg.insertPlainText(str(json.dumps(response, indent=4)))
    #textWdg.insertPlainText(_utils.report_to_ascii(response, "", ""))
    if error:
        textWdg.insertPlainText("{0}\n\nBe sure that there's a rule to allow outbound connections from the GUI to www.virustotal.com".format(
            error
        ))
    else:
        textWdg.setHtml(_utils.report_to_html(response))
    textWdg.moveCursor(QtGui.QTextCursor.Start)

def add_analyzing_msg(vt, parent):
    parent.messageLabel.setText("{0}<br>{1}".format(
        vt.ANALYZING_MESSAGE,
        parent.messageLabel.text()
    ))

def reset_widgets_state(parent):
    parent.messageLabel.setStyleSheet('')
    parent.appNameLabel.setStyleSheet('')
    parent.checksumLabel.setStyleSheet('')
    parent.destIPLabel.setStyleSheet('')

def _cb_popup_link_clicked(link, parent):
    """link clicked on the popup"""
    if link == "#virustotal-warning":
        wdg_count = parent.stackedWidget.count()
        parent.stackedWidget.setCurrentIndex(3)




from PyQt5.QtSql import QSqlQuery

from opensnitch.utils import AsnDB
from opensnitch.customwidgets.generictableview import GenericTableModel
from PyQt5.QtCore import QCoreApplication as QC

class AddressTableModel(GenericTableModel):

    def __init__(self, tableName, headerLabels):
        super().__init__(tableName, headerLabels)
        self.asndb = AsnDB.instance()
        self.reconfigureColumns()

    def reconfigureColumns(self):
        self.headerLabels = []
        self.setHorizontalHeaderLabels(self.headerLabels)
        self.headerLabels.append(QC.translate("stats", "What", ""))
        self.headerLabels.append(QC.translate("stats", "Hits", ""))
        self.headerLabels.append(QC.translate("stats", "Network name", ""))
        self.setHorizontalHeaderLabels(self.headerLabels)
        self.setColumnCount(len(self.headerLabels))
        self.lastColumnCount = len(self.headerLabels)

    def setQuery(self, q, db):
        self.origQueryStr = q
        self.db = db

        if self.prevQueryStr != self.origQueryStr:
            self.realQuery = QSqlQuery(q, db)

        self.realQuery.exec_()
        self.realQuery.last()

        queryRows = max(0, self.realQuery.at()+1)
        self.totalRowCount = queryRows
        self.setRowCount(self.totalRowCount)

        queryColumns = self.realQuery.record().count()
        if self.asndb.is_available() and queryColumns < 3:
            self.reconfigureColumns()
        else:
            # update view's columns
            if queryColumns != self.lastColumnCount:
                self.setModelColumns(queryColumns)

        self.prevQueryStr = self.origQueryStr
        self.rowCountChanged.emit()

    def fillVisibleRows(self, q, upperBound, force=False):
        super().fillVisibleRows(q, upperBound, force)

        if self.asndb.is_available() == True and self.columnCount() <= 3:
            for n, col in enumerate(self.items):
                try:
                    if len(col) < 2:
                        continue
                    col[2] = self.asndb.get_asn(col[0])
                except:
                    col[2] = ""
                finally:
                    self.items[n] = col
            self.lastItems = self.items




from PyQt5 import QtWidgets

# TODO
def add_panel_items(parent, config):
    """add an entry to the Rules Tab -> left panel, to list configured urls"""
    cmdPrefs = QtWidgets.QPushButton("x", objectName="cmdPrefsDownloaders")
    cmdPrefs.setFlat(True)
    #cmdPregs.clicked.connect(_cb_prefs_lists_clicked)
    itemTop = QtWidgets.QTreeWidgetItem(parent.rulesTreePanel)
    font = itemTop.font(0)
    font.setBold(True)
    itemTop.setFont(0, font)
    itemTop.setText(0, "lists")
    for idx, cfg in enumerate(config):
        for url in cfg['urls']:
            item = QtWidgets.QTreeWidgetItem(itemTop)
            item.setText(0, url['name'])

    parent.rulesTreePanel.addTopLevelItem(itemTop)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import ui_pb2 as ui__pb2


class UIStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Ping = channel.unary_unary(
                '/protocol.UI/Ping',
                request_serializer=ui__pb2.PingRequest.SerializeToString,
                response_deserializer=ui__pb2.PingReply.FromString,
                )
        self.AskRule = channel.unary_unary(
                '/protocol.UI/AskRule',
                request_serializer=ui__pb2.Connection.SerializeToString,
                response_deserializer=ui__pb2.Rule.FromString,
                )
        self.Subscribe = channel.unary_unary(
                '/protocol.UI/Subscribe',
                request_serializer=ui__pb2.ClientConfig.SerializeToString,
                response_deserializer=ui__pb2.ClientConfig.FromString,
                )
        self.Notifications = channel.stream_stream(
                '/protocol.UI/Notifications',
                request_serializer=ui__pb2.NotificationReply.SerializeToString,
                response_deserializer=ui__pb2.Notification.FromString,
                )
        self.PostAlert = channel.unary_unary(
                '/protocol.UI/PostAlert',
                request_serializer=ui__pb2.Alert.SerializeToString,
                response_deserializer=ui__pb2.MsgResponse.FromString,
                )


class UIServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AskRule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Notifications(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PostAlert(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UIServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=ui__pb2.PingRequest.FromString,
                    response_serializer=ui__pb2.PingReply.SerializeToString,
            ),
            'AskRule': grpc.unary_unary_rpc_method_handler(
                    servicer.AskRule,
                    request_deserializer=ui__pb2.Connection.FromString,
                    response_serializer=ui__pb2.Rule.SerializeToString,
            ),
            'Subscribe': grpc.unary_unary_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=ui__pb2.ClientConfig.FromString,
                    response_serializer=ui__pb2.ClientConfig.SerializeToString,
            ),
            'Notifications': grpc.stream_stream_rpc_method_handler(
                    servicer.Notifications,
                    request_deserializer=ui__pb2.NotificationReply.FromString,
                    response_serializer=ui__pb2.Notification.SerializeToString,
            ),
            'PostAlert': grpc.unary_unary_rpc_method_handler(
                    servicer.PostAlert,
                    request_deserializer=ui__pb2.Alert.FromString,
                    response_serializer=ui__pb2.MsgResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'protocol.UI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UI(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.UI/Ping',
            ui__pb2.PingRequest.SerializeToString,
            ui__pb2.PingReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AskRule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.UI/AskRule',
            ui__pb2.Connection.SerializeToString,
            ui__pb2.Rule.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.UI/Subscribe',
            ui__pb2.ClientConfig.SerializeToString,
            ui__pb2.ClientConfig.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Notifications(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/protocol.UI/Notifications',
            ui__pb2.NotificationReply.SerializeToString,
            ui__pb2.Notification.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PostAlert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.UI/PostAlert',
            ui__pb2.Alert.SerializeToString,
            ui__pb2.MsgResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from PyQt5.QtCore import QObject, pyqtSignal
from queue import Queue
from datetime import datetime
import time
import json

from opensnitch.database import Database
from opensnitch.config import Config
from opensnitch.utils import NetworkInterfaces
from opensnitch.rules import Rules

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

class Nodes(QObject):
    __instance = None
    nodesUpdated = pyqtSignal(int) # total

    LOG_TAG = "[Nodes]: "
    ONLINE = "\u2713 online"
    OFFLINE = "\u2613 offline"
    WARNING = "\u26a0"

    @staticmethod
    def instance():
        if Nodes.__instance == None:
            Nodes.__instance = Nodes()
        return Nodes.__instance

    def __init__(self):
        QObject.__init__(self)
        self._db = Database.instance()
        self._rules = Rules()
        self._nodes = {}
        self._notifications_sent = {}
        self._interfaces = NetworkInterfaces()

    def count(self):
        return len(self._nodes)

    def add(self, _peer, client_config=None):
        try:
            proto, addr = self.get_addr(_peer)
            peer = proto+":"+addr
            if peer not in self._nodes:
                self._nodes[peer] = {
                        'notifications': Queue(),
                        'online':        True,
                        'last_seen':     datetime.now()
                        }
            else:
                self._nodes[peer]['last_seen'] = datetime.now()

            self._nodes[peer]['online'] = True
            self.add_data(peer, client_config)
            self.update(peer)

            self.nodesUpdated.emit(self.count())

            return self._nodes[peer], peer

        except Exception as e:
            print(self.LOG_TAG, "exception adding/updating node: ", e, "addr:", addr, "config:", client_config)

        return None, None

    def add_data(self, addr, client_config):
        if client_config != None:
            self._nodes[addr]['data'] = self.get_client_config(client_config)
            self.add_fw_config(addr, client_config.systemFirewall)
            self._rules.add_rules(addr, client_config.rules)

    def add_fw_config(self, addr, fwconfig):
        self._nodes[addr]['firewall'] = fwconfig

    def add_fw_rules(self, addr, fwconfig):
        self._nodes[addr]['fwrules'] = fwconfig

    def add_rule(self, time, node, name, description, enabled, precedence, nolog, action, duration, op_type, op_sensitive, op_operand, op_data, created):
        # don't add rule if the user has selected to exclude temporary
        # rules
        if duration in Config.RULES_DURATION_FILTER:
            return

        self._rules.add(time, node, name, description, enabled, precedence, nolog, action, duration, op_type, op_sensitive, op_operand, op_data, created)

    def add_rules(self, addr, rules):
        try:
            self._rules.add_rules(addr, rules)
        except Exception as e:
            print(self.LOG_TAG + " exception adding node to db: ", e)

    def delete_rule(self, rule_name, addr, callback):
        deleted_rule = self._rules.delete(rule_name, addr, callback)
        if deleted_rule == None:
            print(self.LOG_TAG, "error deleting rule", rule_name)
            return None, None

        noti = ui_pb2.Notification(type=ui_pb2.DELETE_RULE, rules=[deleted_rule])
        if addr != None:
            nid = self.send_notification(addr, noti, callback)
        else:
            nid = self.send_notifications(noti, callback)

        return nid, noti

    def delete_rule_by_field(self, field, values):
        return self._rules.delete_by_field(field, values)

    def rule_to_json(self, addr, rule_name):
        return self._rules.rule_to_json(addr, rule_name)

    def export_rule(self, addr, rule_name, outdir):
        return self._rules.export_rule(addr, rule_name, outdir)

    def export_rules(self, addr, outdir):
        return self._rules.export_rules(addr, outdir)

    def import_rules(self, addr=None, rulesdir="", callback=None):
        rules_list = self._rules.import_rules(rulesdir)
        if rules_list == None:
            return None, None, None

        notif = ui_pb2.Notification(
                id=int(str(time.time()).replace(".", "")),
                type=ui_pb2.CHANGE_RULE,
                data="",
                rules=rules_list)

        if addr != None:
            nid = self.send_notification(addr, notif, callback)
        else:
            nid = self.send_notifications(notif, callback)

        return nid, notif, rules_list

    def disable_rule(self, addr, rule_name):
        self._rules.disable(addr, rule_name)

    def update_rule_time(self, time, rule_name, addr):
        self._rules.update_time(time, rule_name, addr)

    def delete_all(self):
        self.send_notifications(None)
        self._nodes = {}
        self.nodesUpdated.emit(self.count())

    def delete(self, peer):
        try:
            proto, addr = self.get_addr(peer)
            addr = "%s:%s" % (proto, addr)
            # Force the node to get one new item from queue,
            # in order to loop and exit.
            self._nodes[addr]['notifications'].put(None)
        except:
            addr = peer

        if addr in self._nodes:
            del self._nodes[addr]
            self.nodesUpdated.emit(self.count())

    def get(self):
        return self._nodes

    def get_node(self, addr):
        try:
            return self._nodes[addr]
        except:
            return None

    def get_nodes(self):
        return self._nodes

    def get_node_config(self, addr):
        try:
            if addr not in self._nodes:
                return None
            return self._nodes[addr]['data'].config
        except Exception as e:
            print(self.LOG_TAG + " exception get_node_config(): ", e)
            return None

    def get_client_config(self, client_config):
        try:
            node_config = json.loads(client_config.config)
            if 'LogLevel' not in node_config:
                node_config['LogLevel'] = 1
                client_config.config = json.dumps(node_config)
        except Exception as e:
            print(self.LOG_TAG, "exception parsing client config", e)

        return client_config

    def get_addr(self, peer):
        try:
            peer = peer.split(":")
            # WA for backward compatibility
            if peer[0] == "unix" and peer[1] == "":
                peer[1] = "/local"
            return peer[0], peer[1]
        except:
            print(self.LOG_TAG, "get_addr() error getting addr:", peer)
            return peer

    def is_connected(self, addr):
        try:
            nd = self.get_node(addr)
            return nd != None or (nd in self._nodes and self._nodes[nd]['online'] == True)
        except:
            return None

    def is_local(self, addr):
        if addr.startswith("unix"):
            return True

        if addr.startswith("ipv4") or addr.startswith("ipv6"):
            ifaces = self._interfaces.list()
            for name in ifaces:
                if ifaces[name] in addr:
                    return True

        return False

    def get_notifications(self):
        notlist = []
        try:
            for c in self._nodes:
                if self._nodes[c]['online'] == False:
                    continue
                if self._nodes[c]['notifications'].empty():
                    continue
                notif = self._nodes[c]['notifications'].get(False)
                if notif != None:
                    self._nodes[c]['notifications'].task_done()
                    notlist.append(notif)
        except Exception as e:
            print(self.LOG_TAG + " exception get_notifications(): ", e)

        return notlist

    def save_node_config(self, addr, config):
        try:
            self._nodes[addr]['data'].config = config
        except Exception as e:
            print(self.LOG_TAG + " exception saving node config: ", e, addr, config)

    def save_nodes_config(self, config):
        try:
            for c in self._nodes:
                self._nodes[c]['data'].config = config
        except Exception as e:
            print(self.LOG_TAG + " exception saving nodes config: ", e, config)

    def change_node_config(self, addr, config, _callback):
        _cfg = json.dumps(config, indent="    ")
        notif = ui_pb2.Notification(
            id=int(str(time.time()).replace(".", "")),
            type=ui_pb2.CHANGE_CONFIG,
            data=_cfg,
            rules=[])
        self.save_node_config(addr, _cfg)
        return self.send_notification(addr, notif, _callback), notif

    def start_interception(self, _addr=None, _callback=None):
        return self.firewall(not_type=ui_pb2.ENABLE_INTERCEPTION, addr=_addr, callback=_callback)

    def stop_interception(self, _addr=None, _callback=None):
        return self.firewall(not_type=ui_pb2.DISABLE_INTERCEPTION, addr=_addr, callback=_callback)

    def firewall(self, not_type=ui_pb2.ENABLE_INTERCEPTION, addr=None, callback=None):
        noti = ui_pb2.Notification(clientName="", serverName="", type=not_type, data="", rules=[])
        if addr == None:
            nid = self.send_notifications(noti, callback)
        else:
            nid = self.send_notification(addr, noti, callback)

        return nid, noti

    def send_notification(self, addr, notification, callback_signal=None):
        try:
            notification.id = int(str(time.time()).replace(".", ""))
            if addr not in self._nodes:
                # FIXME: the reply is sent before we return the notification id
                if callback_signal != None:
                    callback_signal.emit(
                        addr,
                        ui_pb2.NotificationReply(
                            id=notification.id,
                            code=ui_pb2.ERROR,
                            data="node not connected: {0}".format(addr)
                        )
                    )
                return notification.id

            self._notifications_sent[notification.id] = {
                    'callback': callback_signal,
                    'type': notification.type
                    }

            self._nodes[addr]['notifications'].put(notification)
        except Exception as e:
            print(self.LOG_TAG + " exception sending notification: ", e, addr, notification)
            if callback_signal != None:
                callback_signal.emit(
                    addr,
                    ui_pb2.NotificationReply(
                        id=notification.id,
                        code=ui_pb2.ERROR,
                        data="Notification not sent ({0}):<br>{1}".format(addr, e)
                    )
                )

        return notification.id

    def send_notifications(self, notification, callback_signal=None):
        """
        Enqueues a notification to the clients queue.
        It'll be retrieved and delivered by get_notifications
        """
        try:
            notification.id = int(str(time.time()).replace(".", ""))
            for c in self._nodes:
                self._nodes[c]['notifications'].put(notification)
            self._notifications_sent[notification.id] = {
                    'callback': callback_signal,
                    'type': notification.type
                    }
        except Exception as e:
            print(self.LOG_TAG + " exception sending notifications: ", e, notification)

        return notification.id

    def reply_notification(self, addr, reply):
        try:
            if reply == None:
                print(self.LOG_TAG, " reply notification None")
                return

            if reply.id not in self._notifications_sent:
                print(self.LOG_TAG, " reply notification not in the list:", reply.id)
                return

            if self._notifications_sent[reply.id] == None:
                print(self.LOG_TAG, " reply notification body empty:", reply.id)
                return

            if self._notifications_sent[reply.id]['callback'] != None:
                self._notifications_sent[reply.id]['callback'].emit(addr, reply)

            # delete only one-time notifications
            # we need the ID of streaming notifications from the server
            # (monitor_process for example) to keep track of the data sent to us.
            if self._notifications_sent[reply.id]['type'] != ui_pb2.TASK_START:
                del self._notifications_sent[reply.id]
        except Exception as e:
            print(self.LOG_TAG, "notification exception:", e)

    def stop_notifications(self):
        """Send a dummy notification to force Notifications class to exit.
        """
        exit_noti = ui_pb2.Notification(clientName="", serverName="", type=0, data="", rules=[])
        self.send_notifications(exit_noti)

    def update(self, peer, status=ONLINE):
        try:
            proto, addr = self.get_addr(peer)
            self._db.update("nodes",
                    "hostname=?,version=?,last_connection=?,status=?",
                    (
                        self._nodes[proto+":"+addr]['data'].name,
                        self._nodes[proto+":"+addr]['data'].version,
                        datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        status,
                        "{0}:{1}".format(proto, addr)),
                        "addr=?"
                    )
        except Exception as e:
            print(self.LOG_TAG + " exception updating DB: ", e, peer)

    def update_all(self, status=OFFLINE):
        try:
            for peer in self._nodes:
                self._db.update("nodes",
                        "hostname=?,version=?,last_connection=?,status=?",
                        (
                            self._nodes[peer]['data'].name,
                            self._nodes[peer]['data'].version,
                            datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            status,
                            peer),
                            "addr=?"
                        )
        except Exception as e:
            print(self.LOG_TAG + " exception updating nodes: ", e)

    def reset_status(self):
        try:
            self._db.update("nodes", "status=?", (self.OFFLINE,))
        except Exception as e:
            print(self.LOG_TAG + " exception resetting nodes status: ", e)

    def reload_fw(self, addr, fw_config, callback):
        notif = ui_pb2.Notification(
                id=int(str(time.time()).replace(".", "")),
                type=ui_pb2.RELOAD_FW_RULES,
                sysFirewall=fw_config
        )
        nid = self.send_notification(addr, notif, callback)
        return nid, notif



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ui.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08ui.proto\x12\x08protocol\"\xcb\x04\n\x05\x41lert\x12\n\n\x02id\x18\x01 \x01(\x04\x12\"\n\x04type\x18\x02 \x01(\x0e\x32\x14.protocol.Alert.Type\x12&\n\x06\x61\x63tion\x18\x03 \x01(\x0e\x32\x16.protocol.Alert.Action\x12*\n\x08priority\x18\x04 \x01(\x0e\x32\x18.protocol.Alert.Priority\x12\"\n\x04what\x18\x05 \x01(\x0e\x32\x14.protocol.Alert.What\x12\x0e\n\x04text\x18\x06 \x01(\tH\x00\x12!\n\x04proc\x18\x08 \x01(\x0b\x32\x11.protocol.ProcessH\x00\x12$\n\x04\x63onn\x18\t \x01(\x0b\x32\x14.protocol.ConnectionH\x00\x12\x1e\n\x04rule\x18\n \x01(\x0b\x32\x0e.protocol.RuleH\x00\x12\"\n\x06\x66wrule\x18\x0b \x01(\x0b\x32\x10.protocol.FwRuleH\x00\")\n\x08Priority\x12\x07\n\x03LOW\x10\x00\x12\n\n\x06MEDIUM\x10\x01\x12\x08\n\x04HIGH\x10\x02\"(\n\x04Type\x12\t\n\x05\x45RROR\x10\x00\x12\x0b\n\x07WARNING\x10\x01\x12\x08\n\x04INFO\x10\x02\"2\n\x06\x41\x63tion\x12\x08\n\x04NONE\x10\x00\x12\x0e\n\nSHOW_ALERT\x10\x01\x12\x0e\n\nSAVE_TO_DB\x10\x02\"l\n\x04What\x12\x0b\n\x07GENERIC\x10\x00\x12\x10\n\x0cPROC_MONITOR\x10\x01\x12\x0c\n\x08\x46IREWALL\x10\x02\x12\x0e\n\nCONNECTION\x10\x03\x12\x08\n\x04RULE\x10\x04\x12\x0b\n\x07NETLINK\x10\x05\x12\x10\n\x0cKERNEL_EVENT\x10\x06\x42\x06\n\x04\x64\x61ta\"\x19\n\x0bMsgResponse\x12\n\n\x02id\x18\x01 \x01(\x04\"o\n\x05\x45vent\x12\x0c\n\x04time\x18\x01 \x01(\t\x12(\n\nconnection\x18\x02 \x01(\x0b\x32\x14.protocol.Connection\x12\x1c\n\x04rule\x18\x03 \x01(\x0b\x32\x0e.protocol.Rule\x12\x10\n\x08unixnano\x18\x04 \x01(\x03\"\xd3\x06\n\nStatistics\x12\x16\n\x0e\x64\x61\x65mon_version\x18\x01 \x01(\t\x12\r\n\x05rules\x18\x02 \x01(\x04\x12\x0e\n\x06uptime\x18\x03 \x01(\x04\x12\x15\n\rdns_responses\x18\x04 \x01(\x04\x12\x13\n\x0b\x63onnections\x18\x05 \x01(\x04\x12\x0f\n\x07ignored\x18\x06 \x01(\x04\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x07 \x01(\x04\x12\x0f\n\x07\x64ropped\x18\x08 \x01(\x04\x12\x11\n\trule_hits\x18\t \x01(\x04\x12\x13\n\x0brule_misses\x18\n \x01(\x04\x12\x33\n\x08\x62y_proto\x18\x0b \x03(\x0b\x32!.protocol.Statistics.ByProtoEntry\x12\x37\n\nby_address\x18\x0c \x03(\x0b\x32#.protocol.Statistics.ByAddressEntry\x12\x31\n\x07\x62y_host\x18\r \x03(\x0b\x32 .protocol.Statistics.ByHostEntry\x12\x31\n\x07\x62y_port\x18\x0e \x03(\x0b\x32 .protocol.Statistics.ByPortEntry\x12/\n\x06\x62y_uid\x18\x0f \x03(\x0b\x32\x1f.protocol.Statistics.ByUidEntry\x12=\n\rby_executable\x18\x10 \x03(\x0b\x32&.protocol.Statistics.ByExecutableEntry\x12\x1f\n\x06\x65vents\x18\x11 \x03(\x0b\x32\x0f.protocol.Event\x1a.\n\x0c\x42yProtoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x30\n\x0e\x42yAddressEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a-\n\x0b\x42yHostEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a-\n\x0b\x42yPortEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a,\n\nByUidEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x33\n\x11\x42yExecutableEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\">\n\x0bPingRequest\x12\n\n\x02id\x18\x01 \x01(\x04\x12#\n\x05stats\x18\x02 \x01(\x0b\x32\x14.protocol.Statistics\"\x17\n\tPingReply\x12\n\n\x02id\x18\x01 \x01(\x04\"\'\n\tStringInt\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\r\"\x9b\x03\n\x07Process\x12\x0b\n\x03pid\x18\x01 \x01(\x04\x12\x0c\n\x04ppid\x18\x02 \x01(\x04\x12\x0b\n\x03uid\x18\x03 \x01(\x04\x12\x0c\n\x04\x63omm\x18\x04 \x01(\t\x12\x0c\n\x04path\x18\x05 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x06 \x03(\t\x12\'\n\x03\x65nv\x18\x07 \x03(\x0b\x32\x1a.protocol.Process.EnvEntry\x12\x0b\n\x03\x63wd\x18\x08 \x01(\t\x12\x33\n\tchecksums\x18\t \x03(\x0b\x32 .protocol.Process.ChecksumsEntry\x12\x10\n\x08io_reads\x18\n \x01(\x04\x12\x11\n\tio_writes\x18\x0b \x01(\x04\x12\x11\n\tnet_reads\x18\x0c \x01(\x04\x12\x12\n\nnet_writes\x18\r \x01(\x04\x12)\n\x0cprocess_tree\x18\x0e \x03(\x0b\x32\x13.protocol.StringInt\x1a*\n\x08\x45nvEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x30\n\x0e\x43hecksumsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xf3\x03\n\nConnection\x12\x10\n\x08protocol\x18\x01 \x01(\t\x12\x0e\n\x06src_ip\x18\x02 \x01(\t\x12\x10\n\x08src_port\x18\x03 \x01(\r\x12\x0e\n\x06\x64st_ip\x18\x04 \x01(\t\x12\x10\n\x08\x64st_host\x18\x05 \x01(\t\x12\x10\n\x08\x64st_port\x18\x06 \x01(\r\x12\x0f\n\x07user_id\x18\x07 \x01(\r\x12\x12\n\nprocess_id\x18\x08 \x01(\r\x12\x14\n\x0cprocess_path\x18\t \x01(\t\x12\x13\n\x0bprocess_cwd\x18\n \x01(\t\x12\x14\n\x0cprocess_args\x18\x0b \x03(\t\x12\x39\n\x0bprocess_env\x18\x0c \x03(\x0b\x32$.protocol.Connection.ProcessEnvEntry\x12\x45\n\x11process_checksums\x18\r \x03(\x0b\x32*.protocol.Connection.ProcessChecksumsEntry\x12)\n\x0cprocess_tree\x18\x0e \x03(\x0b\x32\x13.protocol.StringInt\x1a\x31\n\x0fProcessEnvEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x37\n\x15ProcessChecksumsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"l\n\x08Operator\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07operand\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t\x12\x11\n\tsensitive\x18\x04 \x01(\x08\x12 \n\x04list\x18\x05 \x03(\x0b\x32\x12.protocol.Operator\"\xb6\x01\n\x04Rule\x12\x0f\n\x07\x63reated\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x04 \x01(\x08\x12\x12\n\nprecedence\x18\x05 \x01(\x08\x12\r\n\x05nolog\x18\x06 \x01(\x08\x12\x0e\n\x06\x61\x63tion\x18\x07 \x01(\t\x12\x10\n\x08\x64uration\x18\x08 \x01(\t\x12$\n\x08operator\x18\t \x01(\x0b\x32\x12.protocol.Operator\"-\n\x0fStatementValues\x12\x0b\n\x03Key\x18\x01 \x01(\t\x12\r\n\x05Value\x18\x02 \x01(\t\"P\n\tStatement\x12\n\n\x02Op\x18\x01 \x01(\t\x12\x0c\n\x04Name\x18\x02 \x01(\t\x12)\n\x06Values\x18\x03 \x03(\x0b\x32\x19.protocol.StatementValues\"5\n\x0b\x45xpressions\x12&\n\tStatement\x18\x01 \x01(\x0b\x32\x13.protocol.Statement\"\xd6\x01\n\x06\x46wRule\x12\r\n\x05Table\x18\x01 \x01(\t\x12\r\n\x05\x43hain\x18\x02 \x01(\t\x12\x0c\n\x04UUID\x18\x03 \x01(\t\x12\x0f\n\x07\x45nabled\x18\x04 \x01(\x08\x12\x10\n\x08Position\x18\x05 \x01(\x04\x12\x13\n\x0b\x44\x65scription\x18\x06 \x01(\t\x12\x12\n\nParameters\x18\x07 \x01(\t\x12*\n\x0b\x45xpressions\x18\x08 \x03(\x0b\x32\x15.protocol.Expressions\x12\x0e\n\x06Target\x18\t \x01(\t\x12\x18\n\x10TargetParameters\x18\n \x01(\t\"\x95\x01\n\x07\x46wChain\x12\x0c\n\x04Name\x18\x01 \x01(\t\x12\r\n\x05Table\x18\x02 \x01(\t\x12\x0e\n\x06\x46\x61mily\x18\x03 \x01(\t\x12\x10\n\x08Priority\x18\x04 \x01(\t\x12\x0c\n\x04Type\x18\x05 \x01(\t\x12\x0c\n\x04Hook\x18\x06 \x01(\t\x12\x0e\n\x06Policy\x18\x07 \x01(\t\x12\x1f\n\x05Rules\x18\x08 \x03(\x0b\x32\x10.protocol.FwRule\"M\n\x08\x46wChains\x12\x1e\n\x04Rule\x18\x01 \x01(\x0b\x32\x10.protocol.FwRule\x12!\n\x06\x43hains\x18\x02 \x03(\x0b\x32\x11.protocol.FwChain\"X\n\x0bSysFirewall\x12\x0f\n\x07\x45nabled\x18\x01 \x01(\x08\x12\x0f\n\x07Version\x18\x02 \x01(\r\x12\'\n\x0bSystemRules\x18\x03 \x03(\x0b\x32\x12.protocol.FwChains\"\xc4\x01\n\x0c\x43lientConfig\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x19\n\x11isFirewallRunning\x18\x04 \x01(\x08\x12\x0e\n\x06\x63onfig\x18\x05 \x01(\t\x12\x10\n\x08logLevel\x18\x06 \x01(\r\x12\x1d\n\x05rules\x18\x07 \x03(\x0b\x32\x0e.protocol.Rule\x12-\n\x0esystemFirewall\x18\x08 \x01(\x0b\x32\x15.protocol.SysFirewall\"\xbb\x01\n\x0cNotification\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x12\n\nclientName\x18\x02 \x01(\t\x12\x12\n\nserverName\x18\x03 \x01(\t\x12\x1e\n\x04type\x18\x04 \x01(\x0e\x32\x10.protocol.Action\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\t\x12\x1d\n\x05rules\x18\x06 \x03(\x0b\x32\x0e.protocol.Rule\x12*\n\x0bsysFirewall\x18\x07 \x01(\x0b\x32\x15.protocol.SysFirewall\"\\\n\x11NotificationReply\x12\n\n\x02id\x18\x01 \x01(\x04\x12-\n\x04\x63ode\x18\x02 \x01(\x0e\x32\x1f.protocol.NotificationReplyCode\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t*\x95\x02\n\x06\x41\x63tion\x12\x08\n\x04NONE\x10\x00\x12\x17\n\x13\x45NABLE_INTERCEPTION\x10\x01\x12\x18\n\x14\x44ISABLE_INTERCEPTION\x10\x02\x12\x13\n\x0f\x45NABLE_FIREWALL\x10\x03\x12\x14\n\x10\x44ISABLE_FIREWALL\x10\x04\x12\x13\n\x0fRELOAD_FW_RULES\x10\x05\x12\x11\n\rCHANGE_CONFIG\x10\x06\x12\x0f\n\x0b\x45NABLE_RULE\x10\x07\x12\x10\n\x0c\x44ISABLE_RULE\x10\x08\x12\x0f\n\x0b\x44\x45LETE_RULE\x10\t\x12\x0f\n\x0b\x43HANGE_RULE\x10\n\x12\r\n\tLOG_LEVEL\x10\x0b\x12\x08\n\x04STOP\x10\x0c\x12\x0e\n\nTASK_START\x10\r\x12\r\n\tTASK_STOP\x10\x0e**\n\x15NotificationReplyCode\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x32\xaf\x02\n\x02UI\x12\x34\n\x04Ping\x12\x15.protocol.PingRequest\x1a\x13.protocol.PingReply\"\x00\x12\x31\n\x07\x41skRule\x12\x14.protocol.Connection\x1a\x0e.protocol.Rule\"\x00\x12=\n\tSubscribe\x12\x16.protocol.ClientConfig\x1a\x16.protocol.ClientConfig\"\x00\x12J\n\rNotifications\x12\x1b.protocol.NotificationReply\x1a\x16.protocol.Notification\"\x00(\x01\x30\x01\x12\x35\n\tPostAlert\x12\x0f.protocol.Alert\x1a\x15.protocol.MsgResponse\"\x00\x42\x35Z3github.com/evilsocket/opensnitch/daemon/ui/protocolb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ui_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z3github.com/evilsocket/opensnitch/daemon/ui/protocol'
  _STATISTICS_BYPROTOENTRY._options = None
  _STATISTICS_BYPROTOENTRY._serialized_options = b'8\001'
  _STATISTICS_BYADDRESSENTRY._options = None
  _STATISTICS_BYADDRESSENTRY._serialized_options = b'8\001'
  _STATISTICS_BYHOSTENTRY._options = None
  _STATISTICS_BYHOSTENTRY._serialized_options = b'8\001'
  _STATISTICS_BYPORTENTRY._options = None
  _STATISTICS_BYPORTENTRY._serialized_options = b'8\001'
  _STATISTICS_BYUIDENTRY._options = None
  _STATISTICS_BYUIDENTRY._serialized_options = b'8\001'
  _STATISTICS_BYEXECUTABLEENTRY._options = None
  _STATISTICS_BYEXECUTABLEENTRY._serialized_options = b'8\001'
  _PROCESS_ENVENTRY._options = None
  _PROCESS_ENVENTRY._serialized_options = b'8\001'
  _PROCESS_CHECKSUMSENTRY._options = None
  _PROCESS_CHECKSUMSENTRY._serialized_options = b'8\001'
  _CONNECTION_PROCESSENVENTRY._options = None
  _CONNECTION_PROCESSENVENTRY._serialized_options = b'8\001'
  _CONNECTION_PROCESSCHECKSUMSENTRY._options = None
  _CONNECTION_PROCESSCHECKSUMSENTRY._serialized_options = b'8\001'
  _ACTION._serialized_start=4153
  _ACTION._serialized_end=4430
  _NOTIFICATIONREPLYCODE._serialized_start=4432
  _NOTIFICATIONREPLYCODE._serialized_end=4474
  _ALERT._serialized_start=23
  _ALERT._serialized_end=610
  _ALERT_PRIORITY._serialized_start=357
  _ALERT_PRIORITY._serialized_end=398
  _ALERT_TYPE._serialized_start=400
  _ALERT_TYPE._serialized_end=440
  _ALERT_ACTION._serialized_start=442
  _ALERT_ACTION._serialized_end=492
  _ALERT_WHAT._serialized_start=494
  _ALERT_WHAT._serialized_end=602
  _MSGRESPONSE._serialized_start=612
  _MSGRESPONSE._serialized_end=637
  _EVENT._serialized_start=639
  _EVENT._serialized_end=750
  _STATISTICS._serialized_start=753
  _STATISTICS._serialized_end=1604
  _STATISTICS_BYPROTOENTRY._serialized_start=1315
  _STATISTICS_BYPROTOENTRY._serialized_end=1361
  _STATISTICS_BYADDRESSENTRY._serialized_start=1363
  _STATISTICS_BYADDRESSENTRY._serialized_end=1411
  _STATISTICS_BYHOSTENTRY._serialized_start=1413
  _STATISTICS_BYHOSTENTRY._serialized_end=1458
  _STATISTICS_BYPORTENTRY._serialized_start=1460
  _STATISTICS_BYPORTENTRY._serialized_end=1505
  _STATISTICS_BYUIDENTRY._serialized_start=1507
  _STATISTICS_BYUIDENTRY._serialized_end=1551
  _STATISTICS_BYEXECUTABLEENTRY._serialized_start=1553
  _STATISTICS_BYEXECUTABLEENTRY._serialized_end=1604
  _PINGREQUEST._serialized_start=1606
  _PINGREQUEST._serialized_end=1668
  _PINGREPLY._serialized_start=1670
  _PINGREPLY._serialized_end=1693
  _STRINGINT._serialized_start=1695
  _STRINGINT._serialized_end=1734
  _PROCESS._serialized_start=1737
  _PROCESS._serialized_end=2148
  _PROCESS_ENVENTRY._serialized_start=2056
  _PROCESS_ENVENTRY._serialized_end=2098
  _PROCESS_CHECKSUMSENTRY._serialized_start=2100
  _PROCESS_CHECKSUMSENTRY._serialized_end=2148
  _CONNECTION._serialized_start=2151
  _CONNECTION._serialized_end=2650
  _CONNECTION_PROCESSENVENTRY._serialized_start=2544
  _CONNECTION_PROCESSENVENTRY._serialized_end=2593
  _CONNECTION_PROCESSCHECKSUMSENTRY._serialized_start=2595
  _CONNECTION_PROCESSCHECKSUMSENTRY._serialized_end=2650
  _OPERATOR._serialized_start=2652
  _OPERATOR._serialized_end=2760
  _RULE._serialized_start=2763
  _RULE._serialized_end=2945
  _STATEMENTVALUES._serialized_start=2947
  _STATEMENTVALUES._serialized_end=2992
  _STATEMENT._serialized_start=2994
  _STATEMENT._serialized_end=3074
  _EXPRESSIONS._serialized_start=3076
  _EXPRESSIONS._serialized_end=3129
  _FWRULE._serialized_start=3132
  _FWRULE._serialized_end=3346
  _FWCHAIN._serialized_start=3349
  _FWCHAIN._serialized_end=3498
  _FWCHAINS._serialized_start=3500
  _FWCHAINS._serialized_end=3577
  _SYSFIREWALL._serialized_start=3579
  _SYSFIREWALL._serialized_end=3667
  _CLIENTCONFIG._serialized_start=3670
  _CLIENTCONFIG._serialized_end=3866
  _NOTIFICATION._serialized_start=3869
  _NOTIFICATION._serialized_end=4056
  _NOTIFICATIONREPLY._serialized_start=4058
  _NOTIFICATIONREPLY._serialized_end=4150
  _UI._serialized_start=4477
  _UI._serialized_end=4780
# @@protoc_insertion_point(module_scope)




from opensnitch.database import Database
from opensnitch.config import Config
from opensnitch.nodes import Nodes

# grpc object
class ClientConfig:
    version = "1.2.3"
    name = "bla"
    logLevel = 0
    isFirewallRunning = False
    rules = []
    config = '''{
    "Server":{
        "Address": "unix:///tmp/osui.sock",
        "LogFile": "/var/log/opensnitchd.log"
    },
    "DefaultAction": "deny",
    "DefaultDuration": "once",
    "InterceptUnknown": false,
    "ProcMonitorMethod": "ebpf",
    "LogLevel": 0,
    "LogUTC": true,
    "LogMicro": false,
    "Firewall": "iptables",
    "Stats": {
        "MaxEvents": 150,
        "MaxStats": 50
    }
    }
    '''

class Connection:
    protocol = "tcp"
    src_ip = "127.0.0.1"
    src_port = "12345"
    dst_ip = "127.0.0.1"
    dst_host = "localhost"
    dst_port = "54321"
    user_id = 1000
    process_id = 9876
    process_path = "/bin/cmd"
    process_cwd = "/tmp"
    process_args = "/bin/cmd --parm1 test"
    process_env = []

db = Database.instance()
db.initialize()
Config.init()

nodes = Nodes.instance()
nodes._nodes["unix:/tmp/osui.sock"] = {
    'data': ClientConfig
}



import requests
import re
import ipaddress
import datetime
import os

lists = ( \
    "https://raw.githubusercontent.com/StevenBlack/hosts/master/hosts",
    "https://mirror1.malwaredomains.com/files/justdomains",
    "https://sysctl.org/cameleon/hosts",
    "https://zeustracker.abuse.ch/blocklist.php?download=domainblocklist",
    "https://s3.amazonaws.com/lists.disconnect.me/simple_tracking.txt",
    "https://s3.amazonaws.com/lists.disconnect.me/simple_ad.txt",
    "https://hosts-file.net/ad_servers.txt" )

domains = {}

for url in lists:
    print "Downloading %s ..." % url
    r = requests.get(url)
    if r.status_code != 200:
        print "Error, status code %d" % r.status_code
        continue

    for line in r.text.split("\n"):
        line = line.strip()
        if line == "":
            continue

        elif line[0] == "#":
            continue

        for part in re.split(r'\s+', line):
            part = part.strip()
            if part == "":
                continue

            try:
                duh = ipaddress.ip_address(part)
            except ValueError:
                if part != "localhost":
                    domains[part] = 1

print "Got %d unique domains, saving as rules to ./rules/ ..." % len(domains)

os.system("mkdir -p rules")

idx = 0
for domain, _ in domains.iteritems():
    with open("rules/adv-%d.json" % idx, "wt") as fp:
        tpl = """
{
   "created": "%s",
   "updated": "%s",
   "name": "deny-adv-%d",
   "enabled": true,
   "action": "deny",
   "duration": "always",
   "operator": {
     "type": "simple",
     "operand": "dest.host",
     "data": "%s"
   }
}"""
        now = datetime.datetime.utcnow().isoformat("T") + "Z"
        data = tpl % ( now, now, idx, domain )
        fp.write(data)

    idx = idx + 1



#
# pytest -v tests/dialogs/test_ruleseditor.py
#
import os
import time
import json
from PyQt5 import QtCore, QtWidgets, QtGui

from opensnitch.config import Config
from opensnitch.dialogs.preferences import PreferencesDialog

class TestPreferences():

    @classmethod
    def reset_settings(self):
        try:
            os.remove(os.environ['HOME'] + "/.config/opensnitch/settings.conf")
        except Exception:
            pass

    @classmethod
    def setup_method(self):
        white_icon = QtGui.QIcon("../res/icon-white.svg")
        self.reset_settings()
        self.prefs = PreferencesDialog(appicon=white_icon)
        self.prefs.show()

    def run(self, qtbot):
        def handle_dialog():
            qtbot.mouseClick(self.prefs.applyButton, QtCore.Qt.LeftButton)
            qtbot.mouseClick(self.prefs.acceptButton, QtCore.Qt.LeftButton)

        QtCore.QTimer.singleShot(500, handle_dialog)
        self.prefs.exec_()

    def test_save_popups_settings(self, qtbot):
        """ Test saving UI related settings.
        """
        qtbot.addWidget(self.prefs)

        self.prefs.comboUIAction.setCurrentIndex(Config.ACTION_ALLOW_IDX)
        self.prefs.comboUITarget.setCurrentIndex(2)
        self.prefs.comboUIDuration.setCurrentIndex(4)
        self.prefs.comboUIDialogPos.setCurrentIndex(2)
        self.prefs.spinUITimeout.setValue(30)
        self.prefs.showAdvancedCheck.setChecked(True)
        self.prefs.uidCheck.setChecked(True)

        self.run(qtbot)

        assert self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_ACTION_KEY) == Config.ACTION_ALLOW_IDX and self.prefs.comboUIAction.currentText() == Config.ACTION_ALLOW
        assert self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_TARGET_KEY) == 2
        assert self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_DURATION_KEY) == 4
        assert self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_TIMEOUT_KEY) == 30
        assert self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_POPUP_POSITION) == 2
        assert self.prefs._cfg.getBool(self.prefs._cfg.DEFAULT_POPUP_ADVANCED) == True
        assert self.prefs._cfg.getBool(self.prefs._cfg.DEFAULT_POPUP_ADVANCED_UID) == True

    def test_save_ui_settings(self, qtbot):
        self.prefs.checkUIRules.setChecked(True)
        self.prefs.comboUIRules.setCurrentIndex(1)
        self.prefs.checkHideNode.setChecked(False)
        self.prefs.checkHideProto.setChecked(False)

        self.run(qtbot)

        assert self.prefs._cfg.getBool(self.prefs._cfg.DEFAULT_IGNORE_RULES) == True and  self.prefs._cfg.getInt(self.prefs._cfg.DEFAULT_IGNORE_TEMPORARY_RULES) == 1
        cols = self.prefs._cfg.getSettings(Config.STATS_SHOW_COLUMNS)
        assert cols == ['0','2','3','5','6']

    def test_save_node_settings(self, qtbot, capsys):
        self.prefs.comboNodeAction.setCurrentIndex(Config.ACTION_ALLOW_IDX)
        self.prefs.comboNodeMonitorMethod.setCurrentIndex(2)
        self.prefs.comboNodeLogLevel.setCurrentIndex(5)
        self.prefs.checkNodeLogUTC.setChecked(False)
        self.prefs.checkNodeLogMicro.setChecked(True)
        self.prefs.checkInterceptUnknown.setChecked(True)
        self.prefs.tabWidget.setCurrentIndex(self.prefs.TAB_NODES)
        self.prefs._node_needs_update = True

        self.run(qtbot)

        assert len(self.prefs._notifications_sent) == 1
        for n in self.prefs._notifications_sent:
            conf = json.loads(self.prefs._notifications_sent[n].data)
            assert conf['InterceptUnknown'] == True
            assert conf['ProcMonitorMethod'] == "audit"
            assert conf['LogLevel'] == 5
            assert conf['LogUTC'] == False
            assert conf['LogMicro'] == True
            assert conf['DefaultAction'] == "allow"

# TODO: click on the QMessageDialog
#
#    def test_save_db_settings(self, qtbot, monkeypatch, capsys):
#        self.prefs.comboDBType.setCurrentIndex(1)
#        self.prefs.dbLabel.setText('/tmp/test.db')
#
#        def handle_dialog():
#            qtbot.mouseClick(self.prefs.applyButton, QtCore.Qt.LeftButton)
#            # after saving the settings, a warning dialog must appear, informing
#            # the user to restart the GUI
#            time.sleep(.5)
#            msgbox = QtWidgets.QApplication.activeModalWidget()
#            try:
#                assert msgbox != None
#                okBtn = msgbox.button(QtWidgets.QMessageBox.Ok)
#                qtbot.mouseClick(okBtn, QtCore.Qt.LeftButton)
#            except Exception as e:
#                print("test_save_db_Settings() exception:", e)
#            qtbot.mouseClick(self.prefs.acceptButton, QtCore.Qt.LeftButton)
#
#        QtCore.QTimer.singleShot(500, handle_dialog)
#        self.prefs.exec_()

#        assert self.prefs._cfg.getInt(Config.DEFAULT_DB_TYPE_KEY) == 1
#        assert self.prefs._cfg.getSettings(Config.DEFAULT_DB_FILE_KEY) == '/tmp/test.db'

    def test_load_ui_settings(self, qtbot, capsys):
        """ reTest saved settings (load_settings()).
        On dialog show up the widgets must be configured properly, with the settings
        configured in previous tests.
        """
        self.prefs.checkUIRules.setChecked(False)
        self.prefs.comboUIRules.setCurrentIndex(0)
        self.prefs.comboUITarget.setCurrentIndex(0)
        self.prefs.comboUIDuration.setCurrentIndex(0)
        self.prefs.checkHideNode.setChecked(True)
        self.prefs.checkHideProto.setChecked(True)

        def handle_dialog():
            qtbot.mouseClick(self.prefs.cancelButton, QtCore.Qt.LeftButton)
        QtCore.QTimer.singleShot(500, handle_dialog)

        self.prefs.exec_()
        self.prefs.show()

        print(self.prefs._cfg.getBool(self.prefs._cfg.DEFAULT_IGNORE_RULES))

        assert self.prefs.comboUIAction.currentIndex() == Config.ACTION_ALLOW_IDX and self.prefs.comboUIAction.currentText() == Config.ACTION_ALLOW
        assert self.prefs.checkUIRules.isChecked() == True
        assert self.prefs.comboUIRules.currentIndex() == 1
        assert self.prefs.comboUITarget.currentIndex() == 2
        assert self.prefs.comboUIDuration.currentIndex() == 4 and self.prefs.comboUIDuration.currentText() == Config.DURATION_30m
        assert self.prefs.comboUIDialogPos.currentIndex() == 2
        assert self.prefs.spinUITimeout.value() == 30
